#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Tests for the file-like object implementation using pyvslvm."""

import os
import unittest

from dfvfs.file_io import lvm_file_io
from dfvfs.lib import definitions
from dfvfs.lib import errors
from dfvfs.path import factory as path_spec_factory
from dfvfs.resolver import context

from tests import test_lib as shared_test_lib
from tests.file_io import test_lib


class LVMFileTest(shared_test_lib.BaseTestCase):
  """Tests the Logical Volume Manager (LVM) file-like object."""

  def setUp(self):
    """Sets up the needed objects used throughout the test."""
    self._resolver_context = context.Context()
    test_path = self._GetTestFilePath(['lvm.raw'])
    self._SkipIfPathNotExists(test_path)

    test_os_path_spec = path_spec_factory.Factory.NewPathSpec(
        definitions.TYPE_INDICATOR_OS, location=test_path)
    self._raw_path_spec = path_spec_factory.Factory.NewPathSpec(
        definitions.TYPE_INDICATOR_RAW, parent=test_os_path_spec)

  def tearDown(self):
    """Cleans up the needed objects used throughout the test."""
    self._resolver_context.Empty()

  def testOpenClose(self):
    """Test the open and close functionality."""
    path_spec = path_spec_factory.Factory.NewPathSpec(
        definitions.TYPE_INDICATOR_LVM, parent=self._raw_path_spec,
        volume_index=0)
    file_object = lvm_file_io.LVMFile(self._resolver_context, path_spec)

    file_object.Open()
    self.assertEqual(file_object.get_size(), 4194304)

    path_spec = path_spec_factory.Factory.NewPathSpec(
        definitions.TYPE_INDICATOR_LVM, parent=self._raw_path_spec,
        volume_index=9)
    file_object = lvm_file_io.LVMFile(self._resolver_context, path_spec)

    with self.assertRaises(errors.PathSpecError):
      file_object.Open()

    path_spec = path_spec_factory.Factory.NewPathSpec(
        definitions.TYPE_INDICATOR_LVM, location='/lvm1',
        parent=self._raw_path_spec)
    file_object = lvm_file_io.LVMFile(self._resolver_context, path_spec)

    file_object.Open()
    self.assertEqual(file_object.get_size(), 4194304)

    path_spec = path_spec_factory.Factory.NewPathSpec(
        definitions.TYPE_INDICATOR_LVM, location='/lvm0',
        parent=self._raw_path_spec)
    file_object = lvm_file_io.LVMFile(self._resolver_context, path_spec)

    with self.assertRaises(errors.PathSpecError):
      file_object.Open()

    path_spec = path_spec_factory.Factory.NewPathSpec(
        definitions.TYPE_INDICATOR_LVM, location='/lvm9',
        parent=self._raw_path_spec)
    file_object = lvm_file_io.LVMFile(self._resolver_context, path_spec)

    with self.assertRaises(errors.PathSpecError):
      file_object.Open()

  def testSeek(self):
    """Test the seek functionality."""
    path_spec = path_spec_factory.Factory.NewPathSpec(
        definitions.TYPE_INDICATOR_LVM, parent=self._raw_path_spec,
        volume_index=0)
    file_object = lvm_file_io.LVMFile(self._resolver_context, path_spec)

    file_object.Open()
    self.assertEqual(file_object.get_size(), 4194304)

    file_object.seek(0x488)
    self.assertEqual(file_object.get_offset(), 0x00000488)
    self.assertEqual(file_object.read(11), b'/mnt/dfvfs\x00')
    self.assertEqual(file_object.get_offset(), 0x00000493)

    file_object.seek(-1047544, os.SEEK_END)
    self.assertEqual(file_object.get_offset(), 0x00300408)
    self.assertEqual(file_object.read(8), b'er,passw')
    self.assertEqual(file_object.get_offset(), 0x00300410)

    file_object.seek(3, os.SEEK_CUR)
    self.assertEqual(file_object.get_offset(), 0x00300413)
    self.assertEqual(file_object.read(7), b'\nbank,j')
    self.assertEqual(file_object.get_offset(), 0x0030041a)

    # Conforming to the POSIX seek the offset can exceed the file size
    # but reading will result in no data being returned.
    expected_offset = 4194304 + 100
    file_object.seek(expected_offset, os.SEEK_SET)
    self.assertEqual(file_object.get_offset(), expected_offset)
    self.assertEqual(file_object.read(20), b'')

    with self.assertRaises(IOError):
      file_object.seek(-10, os.SEEK_SET)

    # On error the offset should not change.
    self.assertEqual(file_object.get_offset(), expected_offset)

    with self.assertRaises(IOError):
      file_object.seek(10, 5)

    # On error the offset should not change.
    self.assertEqual(file_object.get_offset(), expected_offset)

  def testRead(self):
    """Test the read functionality."""
    path_spec = path_spec_factory.Factory.NewPathSpec(
        definitions.TYPE_INDICATOR_LVM, parent=self._raw_path_spec,
        volume_index=0)
    file_object = lvm_file_io.LVMFile(self._resolver_context, path_spec)

    file_object.Open()
    self.assertEqual(file_object.get_size(), 4194304)

    file_object.seek(0x80400)

    expected_data = (
        b'This is a text file.\n\nWe should be able to parse it.\n')
    self.assertEqual(file_object.read(53), expected_data)


class LVMImageFileTest(test_lib.Ext2ImageFileTestCase):
  """Tests the Logical Volume Manager (LVM) file-like object."""

  def setUp(self):
    """Sets up the needed objects used throughout the test."""
    super(LVMImageFileTest, self).setUp()
    test_path = self._GetTestFilePath(['lvm.raw'])
    self._SkipIfPathNotExists(test_path)

    test_os_path_spec = path_spec_factory.Factory.NewPathSpec(
        definitions.TYPE_INDICATOR_OS, location=test_path)
    self._raw_path_spec = path_spec_factory.Factory.NewPathSpec(
        definitions.TYPE_INDICATOR_RAW, parent=test_os_path_spec)
    self._lvm_path_spec = path_spec_factory.Factory.NewPathSpec(
        definitions.TYPE_INDICATOR_LVM, parent=self._raw_path_spec,
        volume_index=0)

  def testOpenCloseInode(self):
    """Test the open and close functionality using an inode."""
    self._TestOpenCloseInode(self._lvm_path_spec)

  def testOpenCloseLocation(self):
    """Test the open and close functionality using a location."""
    self._TestOpenCloseLocation(self._lvm_path_spec)

  def testSeek(self):
    """Test the seek functionality."""
    self._TestSeek(self._lvm_path_spec)

  def testRead(self):
    """Test the read functionality."""
    self._TestRead(self._lvm_path_spec)


if __name__ == '__main__':
  unittest.main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Tests for the VFS attribute interface."""

import unittest

from dfvfs.vfs import attribute

from tests import test_lib as shared_test_lib


class AttributeTest(shared_test_lib.BaseTestCase):
  """Tests the VFS attribute interface."""

  def testTypeIndicator(self):
    """Tests the type_indicator property."""
    test_attribute = attribute.Attribute()
    self.assertIsNotNone(test_attribute)
    self.assertIsNone(test_attribute.type_indicator)


if __name__ == '__main__':
  unittest.main()



# -*- coding: utf-8 -*-
"""The RC4 decrypter implementation."""

from cryptography.hazmat.primitives.ciphers import algorithms
from cryptography.hazmat import backends
from cryptography.hazmat.primitives import ciphers

from dfvfs.encryption import decrypter
from dfvfs.encryption import manager
from dfvfs.lib import definitions
from dfvfs.lib import errors


class RC4Decrypter(decrypter.Decrypter):
  """RC4 decrypter using Cryptography."""

  ENCRYPTION_METHOD = definitions.ENCRYPTION_METHOD_RC4

  def __init__(self, key=None, **kwargs):
    """Initializes a decrypter.

    Args:
      key (Optional[bytes]): key.
      kwargs (dict): keyword arguments depending on the decrypter.

    Raises:
      BackEndError: when the cryptography backend cannot be initialized.
      ValueError: when key is not set.
    """
    if not key:
      raise ValueError('Missing key.')

    algorithm = algorithms.ARC4(key)

    try:
      backend = backends.default_backend()
    except ImportError as exception:
      raise errors.BackEndError(exception)

    cipher = ciphers.Cipher(algorithm, mode=None, backend=backend)

    super(RC4Decrypter, self).__init__(**kwargs)
    self._cipher_context = cipher.decryptor()

  # pylint: disable=unused-argument
  def Decrypt(self, encrypted_data, finalize=False):
    """Decrypts the encrypted data.

    Args:
      encrypted_data (bytes): encrypted data.
      finalize (Optional[bool]): True if the end of data has been reached and
          the cipher context should be finalized.

    Returns:
      tuple[bytes,bytes]: decrypted data and remaining encrypted data.
    """
    decrypted_data = self._cipher_context.update(encrypted_data)
    return decrypted_data, b''


manager.EncryptionManager.RegisterDecrypter(RC4Decrypter)



# -*- coding: utf-8 -*-
"""The EXT file entry implementation."""

from dfdatetime import posix_time as dfdatetime_posix_time

from dfvfs.lib import definitions
from dfvfs.lib import errors
from dfvfs.path import ext_path_spec
from dfvfs.vfs import attribute
from dfvfs.vfs import ext_attribute
from dfvfs.vfs import ext_directory
from dfvfs.vfs import extent
from dfvfs.vfs import file_entry


class EXTFileEntry(file_entry.FileEntry):
  """File system file entry that uses pyfsext."""

  TYPE_INDICATOR = definitions.TYPE_INDICATOR_EXT

  # Mappings of EXT file types to dfVFS file entry types.
  _ENTRY_TYPES = {
      0x1000: definitions.FILE_ENTRY_TYPE_PIPE,
      0x2000: definitions.FILE_ENTRY_TYPE_CHARACTER_DEVICE,
      0x4000: definitions.FILE_ENTRY_TYPE_DIRECTORY,
      0x6000: definitions.FILE_ENTRY_TYPE_BLOCK_DEVICE,
      0x8000: definitions.FILE_ENTRY_TYPE_FILE,
      0xa000: definitions.FILE_ENTRY_TYPE_LINK,
      0xc000: definitions.FILE_ENTRY_TYPE_SOCKET}

  _NANOSECONDS_PER_SECOND = 1000000000

  def __init__(
      self, resolver_context, file_system, path_spec, fsext_file_entry=None,
      is_root=False, is_virtual=False):
    """Initializes a file entry.

    Args:
      resolver_context (Context): resolver context.
      file_system (FileSystem): file system.
      path_spec (PathSpec): path specification.
      fsext_file_entry (Optional[pyfsext.file_entry]): EXT file entry.
      is_root (Optional[bool]): True if the file entry is the root file entry
          of the corresponding file system.
      is_virtual (Optional[bool]): True if the file entry is a virtual file
          entry emulated by the corresponding file system.

    Raises:
      BackEndError: if the pyfsext file entry is missing.
    """
    if not fsext_file_entry:
      fsext_file_entry = file_system.GetEXTFileEntryByPathSpec(path_spec)
    if not fsext_file_entry:
      raise errors.BackEndError('Missing pyfsext file entry.')

    if is_root:
      file_entry_name = ''
    else:
      file_entry_name = fsext_file_entry.name

    # Use the path specification location to determine the file entry name
    # if the file entry was retrieved by inode.
    if file_entry_name is None:
      location = getattr(path_spec, 'location', None)
      if location:
        location_segments = file_system.SplitPath(location)
        if location_segments:
          file_entry_name = location_segments[-1]

    super(EXTFileEntry, self).__init__(
        resolver_context, file_system, path_spec, is_root=is_root,
        is_virtual=is_virtual)
    self._creation_time = fsext_file_entry.get_creation_time_as_integer()
    self._fsext_file_entry = fsext_file_entry
    self._name = file_entry_name

    self.entry_type = self._ENTRY_TYPES.get(
        fsext_file_entry.file_mode & 0xf000, None)

  def _GetAttributes(self):
    """Retrieves the attributes.

    Returns:
      list[Attribute]: attributes.
    """
    if self._attributes is None:
      self._attributes = []

      for fsext_extended_attribute in (
          self._fsext_file_entry.extended_attributes):
        extended_attribute = ext_attribute.EXTExtendedAttribute(
            fsext_extended_attribute)
        self._attributes.append(extended_attribute)

    return self._attributes

  def _GetDirectory(self):
    """Retrieves a directory.

    Returns:
      EXTDirectory: directory or None if not available.
    """
    if self.entry_type != definitions.FILE_ENTRY_TYPE_DIRECTORY:
      return None

    return ext_directory.EXTDirectory(
        self._file_system, self.path_spec, self._fsext_file_entry)

  def _GetLink(self):
    """Retrieves the link.

    Returns:
      str: path of the linked file.
    """
    if self._link is None:
      self._link = self._fsext_file_entry.symbolic_link_target
      if self._link and self._link[0] != self._file_system.PATH_SEPARATOR:
        # TODO: make link absolute.
        self._link = f'/{self._link:s}'

    return self._link

  def _GetStatAttribute(self):
    """Retrieves a stat attribute.

    Returns:
      StatAttribute: a stat attribute or None if not available.
    """
    stat_attribute = attribute.StatAttribute()
    stat_attribute.device_number = self._fsext_file_entry.device_number
    stat_attribute.group_identifier = self._fsext_file_entry.group_identifier
    stat_attribute.inode_number = self._fsext_file_entry.inode_number
    stat_attribute.mode = self._fsext_file_entry.file_mode
    stat_attribute.number_of_links = self._fsext_file_entry.number_of_links
    stat_attribute.owner_identifier = self._fsext_file_entry.owner_identifier
    stat_attribute.size = self._fsext_file_entry.size
    stat_attribute.type = self.entry_type

    return stat_attribute

  def _GetSubFileEntries(self):
    """Retrieves a sub file entries generator.

    Yields:
      EXTFileEntry: a sub file entry.
    """
    if self._directory is None:
      self._directory = self._GetDirectory()

    if self._directory:
      for path_spec in self._directory.entries:
        yield EXTFileEntry(self._resolver_context, self._file_system, path_spec)

  @property
  def access_time(self):
    """dfdatetime.DateTimeValues: access time or None if not available."""
    timestamp = self._fsext_file_entry.get_access_time_as_integer()

    # If creation time is not present (None) the timestamp precision is in
    # seconds.
    if self._creation_time is None:
      timestamp, _ = divmod(timestamp, self._NANOSECONDS_PER_SECOND)
      return dfdatetime_posix_time.PosixTime(timestamp=timestamp)

    return dfdatetime_posix_time.PosixTimeInNanoseconds(timestamp=timestamp)

  @property
  def change_time(self):
    """dfdatetime.DateTimeValues: change time or None if not available."""
    timestamp = self._fsext_file_entry.get_inode_change_time_as_integer()

    # If creation time is not present (None) the timestamp precision is in
    # seconds.
    if self._creation_time is None:
      timestamp, _ = divmod(timestamp, self._NANOSECONDS_PER_SECOND)
      return dfdatetime_posix_time.PosixTime(timestamp=timestamp)

    return dfdatetime_posix_time.PosixTimeInNanoseconds(timestamp=timestamp)

  @property
  def creation_time(self):
    """dfdatetime.DateTimeValues: creation time or None if not available."""
    # Creation time can be None if not present and 0 if not set.
    if not self._creation_time:
      return None

    return dfdatetime_posix_time.PosixTimeInNanoseconds(
        timestamp=self._creation_time)

  @property
  def deletion_time(self):
    """dfdatetime.DateTimeValues: deletion time or None if not available."""
    timestamp = self._fsext_file_entry.get_deletion_time_as_integer()
    # Deletion time can be 0 if not set.
    if not timestamp:
      return None

    return dfdatetime_posix_time.PosixTime(timestamp=timestamp)

  @property
  def name(self):
    """str: name of the file entry, which does not include the full path."""
    return self._name

  @property
  def modification_time(self):
    """dfdatetime.DateTimeValues: modification time or None if not available."""
    timestamp = self._fsext_file_entry.get_modification_time_as_integer()

    # If creation time is not present (None) the timestamp precision is in
    # seconds.
    if self._creation_time is None:
      timestamp, _ = divmod(timestamp, self._NANOSECONDS_PER_SECOND)
      return dfdatetime_posix_time.PosixTime(timestamp=timestamp)

    return dfdatetime_posix_time.PosixTimeInNanoseconds(timestamp=timestamp)

  @property
  def size(self):
    """int: size of the file entry in bytes or None if not available."""
    return self._fsext_file_entry.size

  def GetExtents(self):
    """Retrieves the extents.

    Returns:
      list[Extent]: the extents.
    """
    if self.entry_type != definitions.FILE_ENTRY_TYPE_FILE:
      return []

    extents = []
    for extent_index in range(self._fsext_file_entry.number_of_extents):
      extent_offset, extent_size, extent_flags = (
          self._fsext_file_entry.get_extent(extent_index))

      if extent_flags & 0x1:
        extent_type = definitions.EXTENT_TYPE_SPARSE
      else:
        extent_type = definitions.EXTENT_TYPE_DATA

      data_stream_extent = extent.Extent(
          extent_type=extent_type, offset=extent_offset, size=extent_size)
      extents.append(data_stream_extent)

    return extents

  def GetEXTFileEntry(self):
    """Retrieves the EXT file entry.

    Returns:
      pyfsext.file_entry: EXT file entry.
    """
    return self._fsext_file_entry

  def GetLinkedFileEntry(self):
    """Retrieves the linked file entry, e.g. for a symbolic link.

    Returns:
      EXTFileEntry: linked file entry or None if not available.
    """
    link = self._GetLink()
    if not link:
      return None

    parent_path_spec = getattr(self.path_spec, 'parent', None)
    path_spec = ext_path_spec.EXTPathSpec(
        location=link, parent=parent_path_spec)

    is_root = bool(link == self._file_system.LOCATION_ROOT)

    return EXTFileEntry(
        self._resolver_context, self._file_system, path_spec, is_root=is_root)

  def GetParentFileEntry(self):
    """Retrieves the parent file entry.

    Returns:
      EXTFileEntry: parent file entry or None if not available.
    """
    parent_location = None

    location = getattr(self.path_spec, 'location', None)
    if location is not None:
      parent_location = self._file_system.DirnamePath(location)
      if parent_location == '':
        parent_location = self._file_system.PATH_SEPARATOR

    parent_path_spec = getattr(self.path_spec, 'parent', None)
    path_spec = ext_path_spec.EXTPathSpec(
        location=parent_location, parent=parent_path_spec)

    is_root = bool(parent_location == self._file_system.LOCATION_ROOT)

    return EXTFileEntry(
        self._resolver_context, self._file_system, path_spec, is_root=is_root)



# -*- coding: utf-8 -*-
"""Sphinx build configuration file."""

import os
import sys

from sphinx.ext import apidoc

from docutils import nodes
from docutils import transforms

# Change PYTHONPATH to include dfvfs module and dependencies.
sys.path.insert(0, os.path.abspath('..'))

import dfvfs  # pylint: disable=wrong-import-position

import utils.dependencies  # pylint: disable=wrong-import-position


# -- General configuration ------------------------------------------------

# If your documentation needs a minimal Sphinx version, state it here.
needs_sphinx = '2.0.1'

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    'recommonmark',
    'sphinx.ext.autodoc',
    'sphinx.ext.coverage',
    'sphinx.ext.doctest',
    'sphinx.ext.napoleon',
    'sphinx.ext.viewcode',
    'sphinx_markdown_tables',
    'sphinx_rtd_theme',
]

# We cannot install architecture dependent Python modules on readthedocs,
# therefore we mock most imports.
pip_installed_modules = set()

dependency_helper = utils.dependencies.DependencyHelper(
    dependencies_file=os.path.join('..', 'dependencies.ini'),
    test_dependencies_file=os.path.join('..', 'test_dependencies.ini'))
modules_to_mock = set(dependency_helper.dependencies.keys())
modules_to_mock = modules_to_mock.difference(pip_installed_modules)

autodoc_mock_imports = sorted(modules_to_mock)

# Options for the Sphinx Napoleon extension, which reads Google-style
# docstrings.
napoleon_google_docstring = True
napoleon_numpy_docstring = False
napoleon_include_private_with_doc = False
napoleon_include_special_with_doc = True

# General information about the project.
# pylint: disable=redefined-builtin
project = 'dfVFS'
copyright = 'The dfVFS authors'
version = dfvfs.__version__
release = dfvfs.__version__

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
exclude_patterns = ['_build']

# The master toctree document.
master_doc = 'index'

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = 'sphinx'


# -- Options for HTML output ----------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
html_theme = 'sphinx_rtd_theme'

# Output file base name for HTML help builder.
htmlhelp_basename = 'dfvfsdoc'


# -- Options linkcheck ----------------------------------------------------

linkcheck_ignore = [
]


# -- Code to rewrite links for readthedocs --------------------------------

# This function is a Sphinx core event callback, the format of which is detailed
# here: https://www.sphinx-doc.org/en/master/extdev/appapi.html#events

# pylint: disable=unused-argument
def RunSphinxAPIDoc(app):
  """Runs sphinx-apidoc to auto-generate documentation.

  Args:
    app (sphinx.application.Sphinx): Sphinx application. Required by the
        the Sphinx event callback API.
  """
  current_directory = os.path.abspath(os.path.dirname(__file__))
  module_path = os.path.join(current_directory, '..', 'dfvfs')
  api_directory = os.path.join(current_directory, 'sources', 'api')
  apidoc.main(['-o', api_directory, module_path, '--force'])


class MarkdownLinkFixer(transforms.Transform):
  """Transform definition to parse .md references to internal pages."""

  default_priority = 1000

  _URI_PREFIXES = []

  def _FixLinks(self, node):
    """Corrects links to .md files not part of the documentation.

    Args:
      node (docutils.nodes.Node): docutils node.

    Returns:
      docutils.nodes.Node: docutils node, with correct URIs outside
          of Markdown pages outside the documentation.
    """
    if isinstance(node, nodes.reference) and 'refuri' in node:
      reference_uri = node['refuri']
      for uri_prefix in self._URI_PREFIXES:
        if (reference_uri.startswith(uri_prefix) and not (
            reference_uri.endswith('.asciidoc') or
            reference_uri.endswith('.md'))):
          node['refuri'] = reference_uri + '.md'
          break

    return node

  def _Traverse(self, node):
    """Traverses the document tree rooted at node.

    Args:
      node (docutils.nodes.Node): docutils node.
    """
    self._FixLinks(node)

    for child_node in node.children:
      self._Traverse(child_node)

  # pylint: disable=arguments-differ
  def apply(self):
    """Applies this transform on document tree."""
    self._Traverse(self.document)


# pylint: invalid-name
def setup(app):
  """Called at Sphinx initialization.

  Args:
    app (sphinx.application.Sphinx): Sphinx application.
  """
  # Triggers sphinx-apidoc to generate API documentation.
  app.connect('builder-inited', RunSphinxAPIDoc)
  app.add_config_value(
      'recommonmark_config', {'enable_auto_toc_tree': True}, True)
  app.add_transform(MarkdownLinkFixer)



# -*- coding: utf-8 -*-
"""The path specification resolver."""

from dfvfs.credentials import keychain
from dfvfs.lib import definitions
from dfvfs.lib import errors
from dfvfs.path import path_spec
from dfvfs.resolver import context


class Resolver(object):
  """Path specification resolver."""

  _resolver_context = context.Context()
  _resolver_helpers_manager = None

  key_chain = keychain.KeyChain()

  @classmethod
  def _GetResolverHelper(cls, type_indicator):
    """Retrieves the path specification resolver helper for the specified type.

    Args:
      type_indicator (str): type indicator.

    Returns:
      ResolverHelper: a resolver helper.
    """
    if not cls._resolver_helpers_manager:
      # Delay the import of the resolver helpers manager to prevent circular
      # imports.
      from dfvfs.resolver_helpers import manager  # pylint: disable=import-outside-toplevel

      cls._resolver_helpers_manager = manager.ResolverHelperManager

    return cls._resolver_helpers_manager.GetHelper(type_indicator)

  @classmethod
  def OpenFileEntry(cls, path_spec_object, resolver_context=None):
    """Opens a file entry object defined by path specification.

    Args:
      path_spec_object (PathSpec): path specification.
      resolver_context (Optional[Context]): resolver context, where None
          represents the built in context which is not multi process safe.

    Returns:
      FileEntry: file entry or None if the path specification could not be
          resolved.
    """
    file_system = cls.OpenFileSystem(
        path_spec_object, resolver_context=resolver_context)

    if resolver_context is None:
      resolver_context = cls._resolver_context

    return file_system.GetFileEntryByPathSpec(path_spec_object)

  @classmethod
  def OpenFileObject(cls, path_spec_object, resolver_context=None):
    """Opens a file-like object defined by path specification.

    Args:
      path_spec_object (PathSpec): path specification.
      resolver_context (Optional[Context]): resolver context, where None
          represents the built in context which is not multi process safe.

    Returns:
      FileIO: file-like object or None if the path specification could not
          be resolved.

    Raises:
      BackEndError: if the file object cannot be opened.
      MountPointError: if the mount point specified in the path specification
          does not exist.
      PathSpecError: if the path specification is incorrect.
      TypeError: if the path specification type is unsupported.
    """
    if not isinstance(path_spec_object, path_spec.PathSpec):
      raise TypeError('Unsupported path specification type.')

    if resolver_context is None:
      resolver_context = cls._resolver_context

    if path_spec_object.type_indicator == definitions.TYPE_INDICATOR_MOUNT:
      if path_spec_object.HasParent():
        raise errors.PathSpecError(
            'Unsupported mount path specification with parent.')

      mount_point = getattr(path_spec_object, 'identifier', None)
      if not mount_point:
        raise errors.PathSpecError(
            'Unsupported path specification without mount point identifier.')

      path_spec_object = resolver_context.GetMountPoint(mount_point)
      if not path_spec_object:
        raise errors.MountPointError(f'No such mount point: {mount_point:s}')

    file_object = resolver_context.GetFileObject(path_spec_object)
    if not file_object:
      resolver_helper = cls._GetResolverHelper(path_spec_object.type_indicator)
      file_object = resolver_helper.NewFileObject(
          resolver_context, path_spec_object)

      try:
        file_object.Open()
      except (IOError, ValueError) as exception:
        raise errors.BackEndError(
            f'Unable to open file object with error: {exception!s}')

      resolver_context.CacheFileObject(path_spec_object, file_object)

    return file_object

  @classmethod
  def OpenFileSystem(cls, path_spec_object, resolver_context=None):
    """Opens a file system object defined by path specification.

    Args:
      path_spec_object (PathSpec): path specification.
      resolver_context (Optional[Context]): resolver context, where None
          represents the built in context which is not multi process safe.

    Returns:
      FileSystem: file system or None if the path specification could not
          be resolved or has no file system object.

    Raises:
      AccessError: if the access to open the file system was denied.
      BackEndError: if the file system cannot be opened.
      MountPointError: if the mount point specified in the path specification
          does not exist.
      PathSpecError: if the path specification is incorrect.
      TypeError: if the path specification type is unsupported.
    """
    if not isinstance(path_spec_object, path_spec.PathSpec):
      raise TypeError('Unsupported path specification type.')

    if resolver_context is None:
      resolver_context = cls._resolver_context

    if path_spec_object.type_indicator == definitions.TYPE_INDICATOR_MOUNT:
      if path_spec_object.HasParent():
        raise errors.PathSpecError(
            'Unsupported mount path specification with parent.')

      mount_point = getattr(path_spec_object, 'identifier', None)
      if not mount_point:
        raise errors.PathSpecError(
            'Unsupported path specification without mount point identifier.')

      path_spec_object = resolver_context.GetMountPoint(mount_point)
      if not path_spec_object:
        raise errors.MountPointError(f'No such mount point: {mount_point:s}')

    file_system = resolver_context.GetFileSystem(path_spec_object)
    if not file_system:
      resolver_helper = cls._GetResolverHelper(path_spec_object.type_indicator)
      file_system = resolver_helper.NewFileSystem(
          resolver_context, path_spec_object)

      try:
        file_system.Open()
      except (IOError, ValueError) as exception:
        raise errors.BackEndError(
            f'Unable to open file system with error: {exception!s}')

      resolver_context.CacheFileSystem(path_spec_object, file_system)

    return file_system



# -*- coding: utf-8 -*-
"""The SleuthKit (TSK) attribute implementation."""

import os

from dfvfs.lib import errors
from dfvfs.vfs import attribute


class TSKAttribute(attribute.Attribute):
  """File system attribute that uses pytsk3."""

  def __init__(self, tsk_file, tsk_attribute):
    """Initializes an attribute.

    Args:
      tsk_file (pytsk3.File): TSK file.
      tsk_attribute (pytsk3.Attribute): TSK attribute.

    Raises:
      BackEndError: if the TSK file or attribute is missing.
    """
    if not tsk_file:
      raise errors.BackEndError('Missing TSK file.')

    if not tsk_attribute:
      raise errors.BackEndError('Missing TSK attribute.')

    super(TSKAttribute, self).__init__()
    self._attribute_type = getattr(tsk_attribute.info, 'type', None)
    self._tsk_attribute = tsk_attribute
    self._tsk_file = tsk_file

  @property
  def attribute_type(self):
    """object: attribute type."""
    return self._attribute_type


class TSKExtendedAttribute(TSKAttribute):
  """File system extended attribute that uses pytsk3."""

  def __init__(self, tsk_file, tsk_attribute):
    """Initializes an attribute.

    Args:
      tsk_file (pytsk3.File): TSK file.
      tsk_attribute (pytsk3.Attribute): TSK attribute.
    """
    super(TSKExtendedAttribute, self).__init__(tsk_file, tsk_attribute)
    self._current_offset = 0
    self._identifier = getattr(tsk_attribute.info, 'id', None)
    self._size = getattr(tsk_attribute.info, 'size', None)

  @property
  def name(self):
    """str: name.

    Raises:
      BackEndError: if pytsk3 returns a non UTF-8 formatted name.
    """
    name = getattr(self._tsk_attribute.info, 'name', None)
    if name:
      try:
        # pytsk3 returns an UTF-8 encoded byte string.
        return name.decode('utf8')
      except UnicodeError:
        raise errors.BackEndError(
            'pytsk3 returned a non UTF-8 formatted name.')

    return name

  # Note: that the following functions do not follow the style guide
  # because they are part of the file-like object interface.
  # pylint: disable=invalid-name

  def read(self, size=None):
    """Reads a byte string from the file input/output (IO) object.

    The function will read a byte string of the specified size or
    all of the remaining data if no size was specified.

    Args:
      size (Optional[int]): number of bytes to read, where None is all
          remaining data.

    Returns:
      bytes: data read.

    Raises:
      IOError: if the read failed.
      OSError: if the read failed.
    """
    if size is None:
      size = self._size

    data = self._tsk_file.read_random(
        self._current_offset, size, self._attribute_type, self._identifier)
    self._current_offset += len(data)
    return data

  def seek(self, offset, whence=os.SEEK_SET):
    """Seeks to an offset within the file input/output (IO) object.

    Args:
      offset (int): offset to seek.
      whence (Optional[int]): value that indicates whether offset is an
          absolute or relative position within the file.

    Raises:
      IOError: if the seek failed.
      OSError: if the seek failed.
    """
    if whence == os.SEEK_CUR:
      offset += self._current_offset
    elif whence == os.SEEK_END:
      offset = self._size - offset
    elif whence != os.SEEK_SET:
      raise IOError('Invalid whence')

    if offset < 0:
      raise IOError('Invalid offset')

    self._current_offset = offset

  # get_offset() is preferred above tell() by the libbfio layer used in libyal.
  def get_offset(self):
    """Retrieves the current offset into the file input/output (IO) object.

    Returns:
      int: current offset into the file input/output (IO) object.
    """
    return self._current_offset

  # Pythonesque alias for get_offset().
  def tell(self):
    """Retrieves the current offset into the file input/output (IO) object."""
    return self.get_offset()

  def get_size(self):
    """Retrieves the size of the file input/output (IO) object.

    Returns:
      int: size of the file input/output (IO) object.
    """
    return self._size

  def seekable(self):
    """Determines if a file input/output (IO) object is seekable.

    Returns:
      bool: True since a file IO object provides a seek method.
    """
    return True



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Tests for the triple DES decrypter object."""

import unittest

from dfvfs.encryption import des3_decrypter
from dfvfs.lib import definitions
from dfvfs.lib import errors

from tests.encryption import test_lib


class DES3DecrypterTestCase(test_lib.DecrypterTestCase):
  """Tests for the triple DES decrypter object."""

  _DES3_INITIALIZATION_VECTOR = b'This IV!'
  _DES3_KEY = b'This is a key123'

  def testInitialization(self):
    """Tests the initialization method."""
    # Test missing initialization vector with valid block cipher mode.
    try:
      des3_decrypter.DES3Decrypter(
          cipher_mode=definitions.ENCRYPTION_MODE_ECB, key=self._DES3_KEY)
    except errors.BackEndError:
      raise unittest.SkipTest('missing cryptograpy triple DES support')

    # Test missing arguments.
    with self.assertRaises(ValueError):
      des3_decrypter.DES3Decrypter()

    # Test unsupported block cipher mode.
    with self.assertRaises(ValueError):
      des3_decrypter.DES3Decrypter(
          cipher_mode='bogus', key=self._DES3_KEY)

    # Test missing initialization vector.
    with self.assertRaises(ValueError):
      des3_decrypter.DES3Decrypter(
          cipher_mode=definitions.ENCRYPTION_MODE_CBC, key=self._DES3_KEY)

    # Test incorrect key size.
    with self.assertRaises(ValueError):
      des3_decrypter.DES3Decrypter(
          cipher_mode=definitions.ENCRYPTION_MODE_ECB, key=b'Wrong key size')

    # Test incorrect initialization vector type.
    with self.assertRaises(TypeError):
      des3_decrypter.DES3Decrypter(
          cipher_mode=definitions.ENCRYPTION_MODE_CBC,
          initialization_vector='Wrong IV type', key=self._DES3_KEY)

    # Test incorrect initialization vector size.
    with self.assertRaises(ValueError):
      des3_decrypter.DES3Decrypter(
          cipher_mode=definitions.ENCRYPTION_MODE_CBC,
          initialization_vector=b'Wrong IV size', key=self._DES3_KEY)

  def testDecrypt(self):
    """Tests the Decrypt method."""
    try:
      decrypter = des3_decrypter.DES3Decrypter(
          cipher_mode=definitions.ENCRYPTION_MODE_CBC,
          initialization_vector=self._DES3_INITIALIZATION_VECTOR,
          key=self._DES3_KEY)
    except errors.BackEndError:
      raise unittest.SkipTest('missing cryptograpy triple DES support')

    # Test full decryption.
    expected_decrypted_data = b'This is secret encrypted text!!!'

    decrypted_data, remaining_encrypted_data = decrypter.Decrypt(
        b'e\x86k\t\x01W\xd7d\xe4\xa4\xb3~\x80\xd3\xc3\x7fq{E}:L\n '
        b'.2\xd1\xcf\x8a\xf1\xa0!', finalize=True)

    self.assertEqual(decrypted_data, expected_decrypted_data)
    self.assertEqual(remaining_encrypted_data, b'')

    # Reset decrypter.
    try:
      decrypter = des3_decrypter.DES3Decrypter(
          cipher_mode=definitions.ENCRYPTION_MODE_CBC,
          initialization_vector=self._DES3_INITIALIZATION_VECTOR,
          key=self._DES3_KEY)
    except errors.BackEndError:
      raise unittest.SkipTest('missing cryptograpy triple DES support')

    # Test partial decryption.
    partial_encrypted_data = b'e\x86k\t\x01W\xd7d\xe4\xa4\xb3~\x80'

    decrypted_data, remaining_encrypted_data = decrypter.Decrypt(
        partial_encrypted_data)
    self.assertEqual(decrypted_data, b'')
    self.assertEqual(remaining_encrypted_data, partial_encrypted_data)


if __name__ == '__main__':
  unittest.main()



# -*- coding: utf-8 -*-
"""The NTFS directory implementation."""

from dfvfs.path import ntfs_path_spec
from dfvfs.vfs import directory


class NTFSDirectory(directory.Directory):
  """File system directory that uses pyfsntfs."""

  _FILE_REFERENCE_MFT_ENTRY_BITMASK = 0xffffffffffff

  def __init__(self, file_system, path_spec, fsntfs_file_entry):
    """Initializes a directory.

    Args:
      file_system (FileSystem): file system.
      path_spec (PathSpec): path specification.
      fsntfs_file_entry (pyfsntfs.file_entry): NTFS file entry.
    """
    super(NTFSDirectory, self).__init__(file_system, path_spec)
    self._fsntfs_file_entry = fsntfs_file_entry

  def _EntriesGenerator(self):
    """Retrieves directory entries.

    Since a directory can contain a vast number of entries using
    a generator is more memory efficient.

    Yields:
      NTFSPathSpec: NTFS path specification.
    """
    location = getattr(self.path_spec, 'location', None)

    for fsntfs_sub_file_entry in self._fsntfs_file_entry.sub_file_entries:
      directory_entry = fsntfs_sub_file_entry.name

      # Ignore references to self or parent.
      if directory_entry in ('.', '..'):
        continue

      file_reference = fsntfs_sub_file_entry.file_reference
      directory_entry_mft_entry = (
          file_reference & self._FILE_REFERENCE_MFT_ENTRY_BITMASK)

      if not location or location == self._file_system.PATH_SEPARATOR:
        directory_entry = self._file_system.JoinPath([directory_entry])
      else:
        directory_entry = self._file_system.JoinPath([
            location, directory_entry])

      yield ntfs_path_spec.NTFSPathSpec(
          location=directory_entry,
          mft_attribute=fsntfs_sub_file_entry.name_attribute_index,
          mft_entry=directory_entry_mft_entry, parent=self.path_spec.parent)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Tests for the serializer object implementation using JSON."""

import os
import unittest

from dfvfs.path import os_path_spec
from dfvfs.path import qcow_path_spec
from dfvfs.path import tsk_path_spec
from dfvfs.path import vshadow_path_spec
from dfvfs.serializer import json_serializer as serializer

from tests import test_lib as shared_test_lib


class JsonPathSpecSerializerTest(shared_test_lib.BaseTestCase):
  """Tests for the JSON path specification serializer."""

  def setUp(self):
    """Sets up the needed objects used throughout the test."""
    test_file = self._GetTestFilePath(['ext2.qcow2'])
    self._os_path_spec = os_path_spec.OSPathSpec(location=test_file)
    self._qcow_path_spec = qcow_path_spec.QCOWPathSpec(
        parent=self._os_path_spec)
    self._vshadow_path_spec = vshadow_path_spec.VShadowPathSpec(
        store_index=1, parent=self._qcow_path_spec)
    self._tsk_path_spec = tsk_path_spec.TSKPathSpec(
        inode=16, location='/a_directory/another_file',
        parent=self._vshadow_path_spec)

    self._tsk_path_spec_dict = {
        'inode': 16,
        'location': '/a_directory/another_file',
        'parent': {
            'store_index': 1,
            'parent': {
                'parent': {
                    'location': os.path.abspath(test_file)}
            }
        }
    }

  def testReadAndWriteSerialized(self):
    """Test the ReadSerialized and WriteSerialized function."""
    serialized_path_spec = serializer.JsonPathSpecSerializer.WriteSerialized(
        self._tsk_path_spec)

    self.assertIsNotNone(serialized_path_spec)

    path_spec = serializer.JsonPathSpecSerializer.ReadSerialized(
        serialized_path_spec)

    self.assertIsNotNone(path_spec)

    path_spec_dict = path_spec.CopyToDict()
    self.assertEqual(
        sorted(path_spec_dict.items()),
        sorted(self._tsk_path_spec_dict.items()))


if __name__ == '__main__':
  unittest.main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Tests for the CPIO extracted file-like object."""

import unittest

from dfvfs.file_io import cpio_file_io
from dfvfs.lib import definitions
from dfvfs.path import factory as path_spec_factory
from dfvfs.resolver import context

from tests.file_io import test_lib


class CPIOBinaryFileTest(test_lib.SylogTestCase):
  """The unit test for a CPIO extracted file-like object."""

  def setUp(self):
    """Sets up the needed objects used throughout the test."""
    super(CPIOBinaryFileTest, self).setUp()
    self._resolver_context = context.Context()
    test_path = self._GetTestFilePath(['syslog.bin.cpio'])
    self._SkipIfPathNotExists(test_path)

    test_os_path_spec = path_spec_factory.Factory.NewPathSpec(
        definitions.TYPE_INDICATOR_OS, location=test_path)
    self._cpio_path_spec = path_spec_factory.Factory.NewPathSpec(
        definitions.TYPE_INDICATOR_CPIO, location='/syslog',
        parent=test_os_path_spec)

  def tearDown(self):
    """Cleans up the needed objects used throughout the test."""
    self._resolver_context.Empty()

  def testOpenClosePathSpec(self):
    """Test the open and close functionality using a path specification."""
    file_object = cpio_file_io.CPIOFile(
        self._resolver_context, self._cpio_path_spec)
    file_object.Open()

    self._TestGetSizeFileObject(file_object)

  def testSeek(self):
    """Test the seek functionality."""
    file_object = cpio_file_io.CPIOFile(
        self._resolver_context, self._cpio_path_spec)
    file_object.Open()

    self._TestSeekFileObject(file_object)

  def testRead(self):
    """Test the read functionality."""
    file_object = cpio_file_io.CPIOFile(
        self._resolver_context, self._cpio_path_spec)
    file_object.Open()

    self._TestReadFileObject(file_object)


class CPIOPortableASCIIFileTest(test_lib.SylogTestCase):
  """The unit test for a CPIO extracted file-like object."""

  def setUp(self):
    """Sets up the needed objects used throughout the test."""
    super(CPIOPortableASCIIFileTest, self).setUp()
    self._resolver_context = context.Context()
    test_path = self._GetTestFilePath(['syslog.odc.cpio'])
    self._SkipIfPathNotExists(test_path)

    test_os_path_spec = path_spec_factory.Factory.NewPathSpec(
        definitions.TYPE_INDICATOR_OS, location=test_path)
    self._cpio_path_spec = path_spec_factory.Factory.NewPathSpec(
        definitions.TYPE_INDICATOR_CPIO, location='/syslog',
        parent=test_os_path_spec)

  def tearDown(self):
    """Cleans up the needed objects used throughout the test."""
    self._resolver_context.Empty()

  def testOpenClosePathSpec(self):
    """Test the open and close functionality using a path specification."""
    file_object = cpio_file_io.CPIOFile(
        self._resolver_context, self._cpio_path_spec)
    file_object.Open()

    self._TestGetSizeFileObject(file_object)

  def testSeek(self):
    """Test the seek functionality."""
    file_object = cpio_file_io.CPIOFile(
        self._resolver_context, self._cpio_path_spec)
    file_object.Open()

    self._TestSeekFileObject(file_object)

  def testRead(self):
    """Test the read functionality."""
    file_object = cpio_file_io.CPIOFile(
        self._resolver_context, self._cpio_path_spec)
    file_object.Open()

    self._TestReadFileObject(file_object)


class CPIONewASCIIFileTest(test_lib.SylogTestCase):
  """The unit test for a CPIO extracted file-like object."""

  def setUp(self):
    """Sets up the needed objects used throughout the test."""
    super(CPIONewASCIIFileTest, self).setUp()
    self._resolver_context = context.Context()
    test_path = self._GetTestFilePath(['syslog.newc.cpio'])
    self._SkipIfPathNotExists(test_path)

    test_os_path_spec = path_spec_factory.Factory.NewPathSpec(
        definitions.TYPE_INDICATOR_OS, location=test_path)
    self._cpio_path_spec = path_spec_factory.Factory.NewPathSpec(
        definitions.TYPE_INDICATOR_CPIO, location='/syslog',
        parent=test_os_path_spec)

  def tearDown(self):
    """Cleans up the needed objects used throughout the test."""
    self._resolver_context.Empty()

  def testOpenClosePathSpec(self):
    """Test the open and close functionality using a path specification."""
    file_object = cpio_file_io.CPIOFile(
        self._resolver_context, self._cpio_path_spec)
    file_object.Open()

    self._TestGetSizeFileObject(file_object)

  def testSeek(self):
    """Test the seek functionality."""
    file_object = cpio_file_io.CPIOFile(
        self._resolver_context, self._cpio_path_spec)
    file_object.Open()

    self._TestSeekFileObject(file_object)

  def testRead(self):
    """Test the read functionality."""
    file_object = cpio_file_io.CPIOFile(
        self._resolver_context, self._cpio_path_spec)
    file_object.Open()

    self._TestReadFileObject(file_object)


class CPIONewASCIIFileWithChecksumTest(test_lib.SylogTestCase):
  """The unit test for a CPIO extracted file-like object."""

  def setUp(self):
    """Sets up the needed objects used throughout the test."""
    super(CPIONewASCIIFileWithChecksumTest, self).setUp()
    self._resolver_context = context.Context()
    test_path = self._GetTestFilePath(['syslog.crc.cpio'])
    self._SkipIfPathNotExists(test_path)

    test_os_path_spec = path_spec_factory.Factory.NewPathSpec(
        definitions.TYPE_INDICATOR_OS, location=test_path)
    self._cpio_path_spec = path_spec_factory.Factory.NewPathSpec(
        definitions.TYPE_INDICATOR_CPIO, location='/syslog',
        parent=test_os_path_spec)

  def tearDown(self):
    """Cleans up the needed objects used throughout the test."""
    self._resolver_context.Empty()

  def testOpenClosePathSpec(self):
    """Test the open and close functionality using a path specification."""
    file_object = cpio_file_io.CPIOFile(
        self._resolver_context, self._cpio_path_spec)
    file_object.Open()

    self._TestGetSizeFileObject(file_object)

  def testSeek(self):
    """Test the seek functionality."""
    file_object = cpio_file_io.CPIOFile(
        self._resolver_context, self._cpio_path_spec)
    file_object.Open()

    self._TestSeekFileObject(file_object)

  def testRead(self):
    """Test the read functionality."""
    file_object = cpio_file_io.CPIOFile(
        self._resolver_context, self._cpio_path_spec)
    file_object.Open()

    self._TestReadFileObject(file_object)


if __name__ == '__main__':
  unittest.main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Tests for the data stream implementation using pyfsntfs."""

import unittest

from dfvfs.lib import definitions
from dfvfs.path import factory as path_spec_factory
from dfvfs.resolver import context
from dfvfs.vfs import ntfs_data_stream
from dfvfs.vfs import ntfs_file_system

from tests import test_lib as shared_test_lib


class NTFSDataStreamTest(shared_test_lib.BaseTestCase):
  """Tests the NTFS data stream."""

  def setUp(self):
    """Sets up the needed objects used throughout the test."""
    self._resolver_context = context.Context()
    test_path = self._GetTestFilePath(['ntfs.raw'])
    self._SkipIfPathNotExists(test_path)

    test_os_path_spec = path_spec_factory.Factory.NewPathSpec(
        definitions.TYPE_INDICATOR_OS, location=test_path)
    self._raw_path_spec = path_spec_factory.Factory.NewPathSpec(
        definitions.TYPE_INDICATOR_RAW, parent=test_os_path_spec)
    self._ntfs_path_spec = path_spec_factory.Factory.NewPathSpec(
        definitions.TYPE_INDICATOR_NTFS, location='\\',
        parent=self._raw_path_spec)

    self._file_system = ntfs_file_system.NTFSFileSystem(
        self._resolver_context, self._ntfs_path_spec)
    self._file_system.Open()

  def tearDown(self):
    """Cleans up the needed objects used throughout the test."""
    self._resolver_context.Empty()

  def testName(self):
    """Test the name property."""
    path_spec = path_spec_factory.Factory.NewPathSpec(
        definitions.TYPE_INDICATOR_NTFS, location='\\$UpCase', mft_entry=10,
        parent=self._raw_path_spec)
    file_entry = self._file_system.GetFileEntryByPathSpec(path_spec)
    self.assertIsNotNone(file_entry)

    test_data_stream = ntfs_data_stream.NTFSDataStream(file_entry, None)
    self.assertEqual(test_data_stream.name, '')

    fsntfs_file_entry = file_entry.GetNTFSFileEntry()
    self.assertIsNotNone(fsntfs_file_entry)

    fsntfs_data_stream = fsntfs_file_entry.get_alternate_data_stream_by_name(
        '$Info')

    test_data_stream = ntfs_data_stream.NTFSDataStream(
        file_entry, fsntfs_data_stream)
    self.assertEqual(test_data_stream.name, '$Info')

  def testGetExtents(self):
    """Test the GetExtents function."""
    path_spec = path_spec_factory.Factory.NewPathSpec(
        definitions.TYPE_INDICATOR_NTFS, location='\\$UpCase', mft_entry=10,
        parent=self._raw_path_spec)
    file_entry = self._file_system.GetFileEntryByPathSpec(path_spec)
    self.assertIsNotNone(file_entry)

    test_data_stream = ntfs_data_stream.NTFSDataStream(file_entry, None)

    extents = test_data_stream.GetExtents()
    self.assertEqual(len(extents), 1)

    self.assertEqual(extents[0].extent_type, definitions.EXTENT_TYPE_DATA)
    self.assertEqual(extents[0].offset, 823296)
    self.assertEqual(extents[0].size, 131072)

    fsntfs_file_entry = file_entry.GetNTFSFileEntry()
    self.assertIsNotNone(fsntfs_file_entry)

    fsntfs_data_stream = fsntfs_file_entry.get_alternate_data_stream_by_name(
        '$Info')

    test_data_stream = ntfs_data_stream.NTFSDataStream(
        file_entry, fsntfs_data_stream)

    extents = test_data_stream.GetExtents()
    # No extents are returned for a resident $DATA attribute.
    self.assertEqual(len(extents), 0)

  def testIsDefault(self):
    """Test the IsDefault function."""
    path_spec = path_spec_factory.Factory.NewPathSpec(
        definitions.TYPE_INDICATOR_NTFS, location='\\$UpCase', mft_entry=10,
        parent=self._raw_path_spec)
    file_entry = self._file_system.GetFileEntryByPathSpec(path_spec)
    self.assertIsNotNone(file_entry)

    test_data_stream = ntfs_data_stream.NTFSDataStream(file_entry, None)

    result = test_data_stream.IsDefault()
    self.assertTrue(result)

    fsntfs_file_entry = file_entry.GetNTFSFileEntry()
    self.assertIsNotNone(fsntfs_file_entry)

    fsntfs_data_stream = fsntfs_file_entry.get_alternate_data_stream_by_name(
        '$Info')

    test_data_stream = ntfs_data_stream.NTFSDataStream(
        file_entry, fsntfs_data_stream)

    result = test_data_stream.IsDefault()
    self.assertFalse(result)


if __name__ == '__main__':
  unittest.main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Tests for gzip compressed stream file."""

# Note: do not rename file to gzip.py this can cause the exception:
# AttributeError: 'module' object has no attribute 'GzipFile'
# when using pip.

import os
import unittest

from dfvfs.lib import definitions
from dfvfs.lib import gzipfile
from dfvfs.path import factory as path_spec_factory
from dfvfs.resolver import resolver

from tests import test_lib as shared_test_lib


# TODO: add tests for _GzipDecompressorState
# TODO: add tests for GzipMember


class GzipCompressedStreamTest(shared_test_lib.BaseTestCase):
  """Tests a gzip compressed stream file-like object."""

  def testOpenClose(self):
    """Test the Open and Close functions."""
    test_path = self._GetTestFilePath(['syslog.gz'])
    self._SkipIfPathNotExists(test_path)

    test_file = gzipfile.GzipCompressedStream()

    test_os_path_spec = path_spec_factory.Factory.NewPathSpec(
        definitions.TYPE_INDICATOR_OS, location=test_path)
    file_object = resolver.Resolver.OpenFileObject(test_os_path_spec)

    test_file.Open(file_object)

    try:
      self.assertEqual(len(test_file.members), 1)

    finally:
      test_file.close()

  def testSeek(self):
    """Test the seek functionality."""
    test_path = self._GetTestFilePath(['syslog.gz'])
    self._SkipIfPathNotExists(test_path)

    test_file = gzipfile.GzipCompressedStream()

    test_os_path_spec = path_spec_factory.Factory.NewPathSpec(
        definitions.TYPE_INDICATOR_OS, location=test_path)
    file_object = resolver.Resolver.OpenFileObject(test_os_path_spec)

    test_file.Open(file_object)

    try:
      test_file.seek(177)
      self.assertEqual(test_file.read(5), b'53:01')

      self.assertEqual(test_file.get_offset(), 182)

      test_file.seek(-10, os.SEEK_END)
      self.assertEqual(test_file.read(5), b'times')

      test_file.seek(2, os.SEEK_CUR)
      self.assertEqual(test_file.read(2), b'--')

      # Conforming to the POSIX seek the offset can exceed the file size
      # but reading will result in no data being returned.
      test_file.seek(2000, os.SEEK_SET)
      self.assertEqual(test_file.get_offset(), 2000)
      self.assertEqual(test_file.read(2), b'')

      # Test with an invalid offset.
      with self.assertRaises(IOError):
        test_file.seek(-10, os.SEEK_SET)

      # On error the offset should not change.
      self.assertEqual(test_file.get_offset(), 2000)

      # Test with an invalid whence.
      with self.assertRaises(IOError):
        test_file.seek(10, 5)

      # On error the offset should not change.
      self.assertEqual(test_file.get_offset(), 2000)

    finally:
      test_file.close()

  def testRead(self):
    """Test the read functionality."""
    test_path = self._GetTestFilePath(['syslog.gz'])
    self._SkipIfPathNotExists(test_path)

    test_file = gzipfile.GzipCompressedStream()

    test_os_path_spec = path_spec_factory.Factory.NewPathSpec(
        definitions.TYPE_INDICATOR_OS, location=test_path)
    file_object = resolver.Resolver.OpenFileObject(test_os_path_spec)

    test_file.Open(file_object)

    try:
      test_file.seek(167, os.SEEK_SET)

      self.assertEqual(test_file.get_offset(), 167)

      expected_data = (
          b'Jan 22 07:53:01 myhostname.myhost.com CRON[31051]: (root) CMD '
          b'(touch /var/run/crond.somecheck)\n')

      data = test_file.read(95)
      self.assertEqual(data, expected_data)

      self.assertEqual(test_file.get_offset(), 262)

    finally:
      test_file.close()

  def testReadCorrupt(self):
    """Tests reading a file that is corrupt."""
    # The corrupt gzip has no member footer.
    test_path = self._GetTestFilePath(['corrupt1.gz'])
    self._SkipIfPathNotExists(test_path)

    test_file = gzipfile.GzipCompressedStream()

    test_os_path_spec = path_spec_factory.Factory.NewPathSpec(
        definitions.TYPE_INDICATOR_OS, location=test_path)
    file_object = resolver.Resolver.OpenFileObject(test_os_path_spec)

    test_file.Open(file_object)

    try:
      self.assertEqual(test_file.uncompressed_data_size, 2994187)

    finally:
      test_file.close()

  def testReadMultipleMembers(self):
    """Tests reading a file that contains multiple gzip members."""
    test_path = self._GetTestFilePath(['fsevents_000000000000b208'])
    self._SkipIfPathNotExists(test_path)

    test_file = gzipfile.GzipCompressedStream()

    test_os_path_spec = path_spec_factory.Factory.NewPathSpec(
        definitions.TYPE_INDICATOR_OS, location=test_path)
    file_object = resolver.Resolver.OpenFileObject(test_os_path_spec)

    test_file.Open(file_object)

    try:
      self.assertEqual(len(test_file.members), 2)

      file_start = test_file.read(4)
      self.assertEqual(file_start, b'1SLD')

      # Read the end of the second member
      test_file.seek(506631 - 4)
      file_end = test_file.read(4)
      self.assertEqual(file_end, b'\x02\x00\x80\x00')

      # Seek backwards, and read across a member boundary.
      test_file.seek(28530)
      self.assertEqual(test_file.read(6), b'OS\x00P\x07\x00')

      # Read with a size greater than the file size.
      test_file.seek(0)
      data = test_file.read(size=506631 + 4)
      self.assertEqual(len(data), 506631)
      self.assertEqual(data[-4:], b'\x02\x00\x80\x00')

    finally:
      test_file.close()


if __name__ == '__main__':
  unittest.main()



# -*- coding: utf-8 -*-
"""Imports for the volume system factory."""

from dfvfs.volume import apfs_volume_system
from dfvfs.volume import cs_volume_system
from dfvfs.volume import gpt_volume_system
from dfvfs.volume import lvm_volume_system
from dfvfs.volume import tsk_volume_system
from dfvfs.volume import vshadow_volume_system



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Tests for the SleuthKit (TSK) volume system."""

import unittest

from dfvfs.lib import definitions
from dfvfs.path import factory as path_spec_factory
from dfvfs.volume import tsk_volume_system

from tests import test_lib as shared_test_lib


class TSKVolumeSystemTestAPM(shared_test_lib.BaseTestCase):
  """Tests the SleuthKit (TSK) volume system on APM."""

  def setUp(self):
    """Sets up the needed objects used throughout the test."""
    test_path = self._GetTestFilePath(['apm.dmg'])
    self._SkipIfPathNotExists(test_path)

    test_os_path_spec = path_spec_factory.Factory.NewPathSpec(
        definitions.TYPE_INDICATOR_OS, location=test_path)
    self._tsk_path_spec = path_spec_factory.Factory.NewPathSpec(
        definitions.TYPE_INDICATOR_TSK_PARTITION, location='/',
        parent=test_os_path_spec)

  # mmls test_data/apm.dmg
  # MAC Partition Map
  # Offset Sector: 0
  # Units are in 512-byte sectors
  #
  #       Slot      Start        End          Length       Description
  # 000:  -------   0000000000   0000000000   0000000001   Unallocated
  # 001:  000       0000000001   0000000063   0000000063   Apple_partition_map
  # 002:  Meta      0000000001   0000000003   0000000003   Table
  # 003:  001       0000000064   0000008175   0000008112   Apple_HFS
  # 004:  002       0000008176   0000008191   0000000016   Apple_Free

  def testIterateVolumes(self):
    """Test the iterate volumes functionality."""
    volume_system = tsk_volume_system.TSKVolumeSystem()
    volume_system.Open(self._tsk_path_spec)

    self.assertEqual(volume_system.bytes_per_sector, 512)

    self.assertEqual(volume_system.number_of_sections, 5)
    self.assertEqual(volume_system.number_of_volumes, 1)

    self.assertEqual(volume_system.volume_identifiers, ['p1'])

    volume = volume_system.GetVolumeByIndex(0)
    self.assertIsNotNone(volume)

    self.assertEqual(volume.number_of_extents, 1)
    self.assertEqual(volume.number_of_attributes, 2)
    self.assertEqual(volume.identifier, 'p1')

    volume_attribute = volume.GetAttribute('address')
    self.assertIsNotNone(volume_attribute)
    self.assertEqual(volume_attribute.value, 3)

    volume_attribute = volume.GetAttribute('description')
    self.assertIsNotNone(volume_attribute)
    self.assertEqual(volume_attribute.value, 'Apple_HFS')

    volume_extent = volume.extents[0]
    self.assertIsNotNone(volume_extent)
    self.assertEqual(volume_extent.offset, 64 * 512)
    self.assertEqual(volume_extent.size, 8112 * 512)
    self.assertEqual(volume_extent.extent_type, volume_extent.EXTENT_TYPE_DATA)

    volume = volume_system.GetVolumeByIndex(9)
    self.assertIsNone(volume)


class TSKVolumeSystemTestGPT(shared_test_lib.BaseTestCase):
  """Tests the SleuthKit (TSK) volume system on GPT."""

  def setUp(self):
    """Sets up the needed objects used throughout the test."""
    test_path = self._GetTestFilePath(['gpt.raw'])
    self._SkipIfPathNotExists(test_path)

    test_os_path_spec = path_spec_factory.Factory.NewPathSpec(
        definitions.TYPE_INDICATOR_OS, location=test_path)
    self._tsk_path_spec = path_spec_factory.Factory.NewPathSpec(
        definitions.TYPE_INDICATOR_TSK_PARTITION, location='/',
        parent=test_os_path_spec)

  # mmls test_data/gpt.raw
  # GUID Partition Table (EFI)
  # Offset Sector: 0
  # Units are in 512-byte sectors
  #
  #       Slot      Start        End          Length       Description
  # 000:  Meta      0000000000   0000000000   0000000001   Safety Table
  # 001:  -------   0000000000   0000002047   0000002048   Unallocated
  # 002:  Meta      0000000001   0000000001   0000000001   GPT Header
  # 003:  Meta      0000000002   0000000033   0000000032   Partition Table
  # 004:  000       0000002048   0000002175   0000000128   Linux filesystem
  # 005:  -------   0000002176   0000004095   0000001920   Unallocated
  # 006:  001       0000004096   0000004223   0000000128   Linux filesystem
  # 007:  -------   0000004224   0000008191   0000003968   Unallocated

  def testIterateVolumes(self):
    """Test the iterate volumes functionality."""
    volume_system = tsk_volume_system.TSKVolumeSystem()
    volume_system.Open(self._tsk_path_spec)

    self.assertEqual(volume_system.bytes_per_sector, 512)

    self.assertEqual(volume_system.number_of_sections, 8)
    self.assertEqual(volume_system.number_of_volumes, 2)

    self.assertEqual(volume_system.volume_identifiers, ['p1', 'p2'])

    volume = volume_system.GetVolumeByIndex(1)
    self.assertIsNotNone(volume)

    self.assertEqual(volume.number_of_extents, 1)
    self.assertEqual(volume.number_of_attributes, 2)
    self.assertEqual(volume.identifier, 'p2')

    volume_attribute = volume.GetAttribute('address')
    self.assertIsNotNone(volume_attribute)
    self.assertEqual(volume_attribute.value, 6)

    volume_attribute = volume.GetAttribute('description')
    self.assertIsNotNone(volume_attribute)
    self.assertEqual(volume_attribute.value, 'Linux filesystem')

    volume_extent = volume.extents[0]
    self.assertIsNotNone(volume_extent)
    self.assertEqual(volume_extent.offset, 4096 * 512)
    self.assertEqual(volume_extent.size, 128 * 512)
    self.assertEqual(volume_extent.extent_type, volume_extent.EXTENT_TYPE_DATA)

    volume = volume_system.GetVolumeByIndex(9)
    self.assertIsNone(volume)


class TSKVolumeSystemTestMBR(shared_test_lib.BaseTestCase):
  """Tests the SleuthKit (TSK) volume system on MBR."""

  def setUp(self):
    """Sets up the needed objects used throughout the test."""
    test_path = self._GetTestFilePath(['mbr.raw'])
    self._SkipIfPathNotExists(test_path)

    test_os_path_spec = path_spec_factory.Factory.NewPathSpec(
        definitions.TYPE_INDICATOR_OS, location=test_path)
    self._tsk_path_spec = path_spec_factory.Factory.NewPathSpec(
        definitions.TYPE_INDICATOR_TSK_PARTITION, location='/',
        parent=test_os_path_spec)

  # mmls test_data/mbr.raw
  # DOS Partition Table
  # Offset Sector: 0
  # Units are in 512-byte sectors
  #
  #       Slot      Start        End          Length       Description
  # 000:  Meta      0000000000   0000000000   0000000001   Primary Table (#0)
  # 001:  -------   0000000000   0000000000   0000000001   Unallocated
  # 002:  000:000   0000000001   0000000129   0000000129   Linux (0x83)
  # 003:  Meta      0000000130   0000008191   0000008062   DOS Extended (0x05)
  # 004:  Meta      0000000130   0000000130   0000000001   Extended Table (#1)
  # 005:  -------   0000000130   0000000130   0000000001   Unallocated
  # 006:  001:000   0000000131   0000000259   0000000129   Linux (0x83)
  # 007:  -------   0000000260   0000008191   0000007932   Unallocated

  def testIterateVolumes(self):
    """Test the iterate volumes functionality."""
    volume_system = tsk_volume_system.TSKVolumeSystem()
    volume_system.Open(self._tsk_path_spec)

    self.assertEqual(volume_system.bytes_per_sector, 512)

    self.assertEqual(volume_system.number_of_sections, 8)
    self.assertEqual(volume_system.number_of_volumes, 2)

    self.assertEqual(volume_system.volume_identifiers, ['p1', 'p2'])

    volume = volume_system.GetVolumeByIndex(1)
    self.assertIsNotNone(volume)

    self.assertEqual(volume.number_of_extents, 1)
    self.assertEqual(volume.number_of_attributes, 2)
    self.assertEqual(volume.identifier, 'p2')

    volume_attribute = volume.GetAttribute('address')
    self.assertIsNotNone(volume_attribute)
    self.assertEqual(volume_attribute.value, 6)

    volume_attribute = volume.GetAttribute('description')
    self.assertIsNotNone(volume_attribute)
    self.assertEqual(volume_attribute.value, 'Linux (0x83)')

    volume_extent = volume.extents[0]
    self.assertIsNotNone(volume_extent)
    self.assertEqual(volume_extent.offset, 131 * 512)
    self.assertEqual(volume_extent.size, 129 * 512)
    self.assertEqual(volume_extent.extent_type, volume_extent.EXTENT_TYPE_DATA)

    volume = volume_system.GetVolumeByIndex(9)
    self.assertIsNone(volume)


if __name__ == '__main__':
  unittest.main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Tests for the VFS directory interface."""

import unittest

from dfvfs.path import fake_path_spec
from dfvfs.resolver import context
from dfvfs.vfs import directory
from dfvfs.vfs import fake_file_system

from tests import test_lib as shared_test_lib


class DirectoryTest(shared_test_lib.BaseTestCase):
  """Tests the VFS directory interface."""

  # pylint: disable=protected-access

  def setUp(self):
    """Sets up the needed objects used throughout the test."""
    self._resolver_context = context.Context()
    self._path_spec = fake_path_spec.FakePathSpec(location='/')

    self._file_system = fake_file_system.FakeFileSystem(
        self._resolver_context, self._path_spec)
    self._file_system.Open()

  def tearDown(self):
    """Cleans up the needed objects used throughout the test."""
    self._resolver_context.Empty()

  def testEntriesGenerator(self):
    """Tests the _EntriesGenerator function."""
    test_directory = directory.Directory(self._file_system, self._path_spec)

    generator = test_directory._EntriesGenerator()
    self.assertIsNotNone(generator)

  def testEntries(self):
    """Tests the entries property."""
    test_directory = directory.Directory(self._file_system, self._path_spec)

    self.assertEqual(list(test_directory.entries), [])


if __name__ == '__main__':
  unittest.main()



# -*- coding: utf-8 -*-
"""The GUID Partition Table (GPT) path specification implementation."""

from dfvfs.lib import definitions
from dfvfs.path import factory
from dfvfs.path import path_spec


class GPTPathSpec(path_spec.PathSpec):
  """GPT path specification.

  Attributes:
    entry_index (int): partition entry index.
    location (str): location.
  """

  TYPE_INDICATOR = definitions.TYPE_INDICATOR_GPT

  def __init__(self, location=None, entry_index=None, parent=None, **kwargs):
    """Initializes a path specification.

    Note that the GPT path specification must have a parent.

    Args:
      entry_index (Optional[int]): partition entry index.
      location (Optional[str]): location.
      parent (Optional[PathSpec]): parent path specification.

    Raises:
      ValueError: when parent is not set.
    """
    if not parent:
      raise ValueError('Missing parent value.')

    super(GPTPathSpec, self).__init__(parent=parent, **kwargs)
    self.entry_index = entry_index
    self.location = location

  @property
  def comparable(self):
    """str: comparable representation of the path specification."""
    string_parts = []

    if self.entry_index is not None:
      string_parts.append(f'entry index: {self.entry_index:d}')
    if self.location is not None:
      string_parts.append(f'location: {self.location:s}')

    return self._GetComparable(sub_comparable_string=', '.join(string_parts))


factory.Factory.RegisterPathSpec(GPTPathSpec)



# -*- coding: utf-8 -*-
"""The operating system directory implementation."""

import os

from dfvfs.lib import errors
from dfvfs.path import os_path_spec
from dfvfs.vfs import directory


class OSDirectory(directory.Directory):
  """File system directory that uses the operating system."""

  def _EntriesGenerator(self):
    """Retrieves directory entries.

    Since a directory can contain a vast number of entries using
    a generator is more memory efficient.

    Yields:
      OSPathSpec: a path specification.

    Raises:
      AccessError: if the access to list the directory was denied.
      BackEndError: if the directory could not be listed.
    """
    location = getattr(self.path_spec, 'location', None)
    if location is not None:
      # Windows will raise WindowsError, which can be caught by OSError,
      # if the process has no access to list the directory. The os.access()
      # function cannot be used since it will return true even when os.listdir()
      # fails.
      try:
        for directory_entry in os.listdir(location):
          directory_entry_location = self._file_system.JoinPath([
              location, directory_entry])
          yield os_path_spec.OSPathSpec(location=directory_entry_location)

      # Note that PermissionError needs to be defined before OSError.
      except PermissionError as exception:
        raise errors.AccessError((
            f'Access to directory: {location:s} denied with error: '
            f'{exception!s}'))

      except OSError as exception:
        raise errors.BackEndError((
            f'Unable to list directory: {location:s} with error: '
            f'{exception!s}'))



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Tests for the EXT attribute."""

import unittest

from dfvfs.lib import definitions
from dfvfs.lib import errors
from dfvfs.path import factory as path_spec_factory
from dfvfs.resolver import context
from dfvfs.vfs import ext_attribute
from dfvfs.vfs import ext_file_system

from tests import test_lib as shared_test_lib


class EXTExtendedAttributeTest(shared_test_lib.BaseTestCase):
  """Tests the EXT extended attribute."""

  # pylint: disable=protected-access

  _INODE_A_FILE = 13

  def setUp(self):
    """Sets up the needed objects used throughout the test."""
    self._resolver_context = context.Context()
    test_path = self._GetTestFilePath(['ext2.raw'])
    self._SkipIfPathNotExists(test_path)

    test_os_path_spec = path_spec_factory.Factory.NewPathSpec(
        definitions.TYPE_INDICATOR_OS, location=test_path)
    self._raw_path_spec = path_spec_factory.Factory.NewPathSpec(
        definitions.TYPE_INDICATOR_RAW, parent=test_os_path_spec)
    self._ext_path_spec = path_spec_factory.Factory.NewPathSpec(
        definitions.TYPE_INDICATOR_EXT, location='/',
        parent=self._raw_path_spec)

    self._file_system = ext_file_system.EXTFileSystem(
        self._resolver_context, self._ext_path_spec)
    self._file_system.Open()

  def tearDown(self):
    """Cleans up the needed objects used throughout the test."""
    self._resolver_context.Empty()

  def testIntialize(self):
    """Tests the __init__ function."""
    test_location = '/a_directory/a_file'
    path_spec = path_spec_factory.Factory.NewPathSpec(
        definitions.TYPE_INDICATOR_EXT, inode=self._INODE_A_FILE,
        location=test_location, parent=self._raw_path_spec)
    file_entry = self._file_system.GetFileEntryByPathSpec(path_spec)

    fsext_attribute = file_entry._fsext_file_entry.get_extended_attribute(0)
    self.assertIsNotNone(fsext_attribute)
    self.assertEqual(fsext_attribute.name, 'user.myxattr')

    test_attribute = ext_attribute.EXTExtendedAttribute(fsext_attribute)
    self.assertIsNotNone(test_attribute)

    with self.assertRaises(errors.BackEndError):
      ext_attribute.EXTExtendedAttribute(None)


if __name__ == '__main__':
  unittest.main()



# -*- coding: utf-8 -*-
"""dtFabric data format helpers."""

import os

from dtfabric import errors as dtfabric_errors
from dtfabric.runtime import data_maps as dtfabric_data_maps

from dfvfs.lib import errors


class DataFormat(object):
  """Data format."""

  def _ReadData(self, file_object, file_offset, data_size):
    """Reads data.

    Args:
      file_object (dfvfs.FileIO): a file-like object to read.
      file_offset (int): offset of the data relative to the start of
          the file-like object.
      data_size (int): size of the data. The resulting data size much match
          the requested data size so that dtFabric can map the data type
          definitions onto the byte stream.

    Returns:
      bytes: byte stream containing the data.

    Raises:
      FileFormatError: if the data cannot be read.
      ValueError: if the file-like object is missing.
    """
    if not file_object:
      raise ValueError('Missing file-like object.')

    file_object.seek(file_offset, os.SEEK_SET)

    read_error = ''

    try:
      data = file_object.read(data_size)

      if len(data) != data_size:
        read_error = 'missing data'

    except IOError as exception:
      read_error = f'{exception!s}'

    if read_error:
      raise errors.FileFormatError((
          f'Unable to read data at offset: 0x{file_offset:08x} with error: '
          f'{read_error:s}'))

    return data

  def _ReadStructureFromFileObject(
      self, file_object, file_offset, data_type_map):
    """Reads a structure from a file-like object.

    If the data type map has a fixed size this method will read the predefined
    number of bytes from the file-like object. If the data type map has a
    variable size, depending on values in the byte stream, this method will
    continue to read from the file-like object until the data type map can be
    successfully mapped onto the byte stream or until an error occurs.

    Args:
      file_object (dfvfs.FileIO): a file-like object to parse.
      file_offset (int): offset of the structure data relative to the start
          of the file-like object.
      data_type_map (dtfabric.DataTypeMap): data type map of the structure.

    Returns:
      tuple[object, int]: structure values object and data size of
          the structure.

    Raises:
      FileFormatError: if the structure cannot be read.
      ValueError: if file-like object or data type map is missing.
    """
    context = None
    data = b''
    last_data_size = 0

    data_size = data_type_map.GetSizeHint()
    while data_size != last_data_size:
      read_offset = file_offset + last_data_size
      read_size = data_size - last_data_size
      data_segment = self._ReadData(file_object, read_offset, read_size)

      data = b''.join([data, data_segment])

      try:
        context = dtfabric_data_maps.DataTypeMapContext()
        structure_values_object = data_type_map.MapByteStream(
            data, context=context)
        return structure_values_object, data_size

      except dtfabric_errors.ByteStreamTooSmallError:
        pass

      except dtfabric_errors.MappingError as exception:
        raise errors.FileFormatError((
            f'Unable to map {data_type_map.name:s} data at offset: '
            f'0x{file_offset:08x} with error: {exception!s}'))

      last_data_size = data_size
      data_size = data_type_map.GetSizeHint(context=context)

    raise errors.FileFormatError((
        f'Unable to read {data_type_map.name:s} at offset: '
        f'0x{file_offset:08x}'))



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Tests for a file system implementation using pyfvde."""

import unittest

from dfvfs.lib import definitions
from dfvfs.path import factory as path_spec_factory
from dfvfs.resolver import context
from dfvfs.vfs import cs_file_system

from tests import test_lib as shared_test_lib


class CSFileSystemTest(shared_test_lib.BaseTestCase):
  """Tests the CS file system."""

  def setUp(self):
    """Sets up the needed objects used throughout the test."""
    self._resolver_context = context.Context()

    test_path = self._GetTestFilePath(['fvdetest.qcow2'])
    self._SkipIfPathNotExists(test_path)

    test_os_path_spec = path_spec_factory.Factory.NewPathSpec(
        definitions.TYPE_INDICATOR_OS, location=test_path)
    test_qcow_path_spec = path_spec_factory.Factory.NewPathSpec(
        definitions.TYPE_INDICATOR_QCOW, parent=test_os_path_spec)
    self._gpt_path_spec = path_spec_factory.Factory.NewPathSpec(
        definitions.TYPE_INDICATOR_GPT, location='/p1',
        parent=test_qcow_path_spec)
    self._cs_path_spec = path_spec_factory.Factory.NewPathSpec(
        definitions.TYPE_INDICATOR_CS, parent=self._gpt_path_spec)

  def tearDown(self):
    """Cleans up the needed objects used throughout the test."""
    self._resolver_context.Empty()

  # fvdeinfo -o $(( 40 * 512 )) fuse/qcow1
  # fvdeinfo 20220120
  #
  # Logical volume: 1 is locked and a password is needed to unlock it.
  #
  # Password:
  #
  # Core Storage information:
  #
  # Logical volume group:
  #     Identifier                  : 94923b58-9f31-4988-8707-cb90c8e45a46
  #     Name                        : TESTLVG
  #     Number of physical volumes  : 1
  #     Number of logical volumes   : 1
  #
  # Physical volume: 1
  #     Identifier                  : 3273a055-3b8b-47e8-b970-df35eecda81b
  #     Size                        : 511 MiB (536829952 bytes)
  #     Encryption method           : AES-XTS
  #
  # Logical volume: 1
  #     Identifier                  : 420af122-cf73-4a30-8b0a-a593a65fbef5
  #     Name                        : TestLV
  #     Size                        : 160 MiB (167772160 bytes)

  def testOpenAndClose(self):
    """Test the open and close functionality."""
    file_system = cs_file_system.CSFileSystem(
        self._resolver_context, self._cs_path_spec)
    self.assertIsNotNone(file_system)

    file_system.Open()

  def testFileEntryExistsByPathSpec(self):
    """Test the file entry exists by path specification functionality."""
    file_system = cs_file_system.CSFileSystem(
        self._resolver_context, self._cs_path_spec)
    self.assertIsNotNone(file_system)

    file_system.Open()

    path_spec = path_spec_factory.Factory.NewPathSpec(
        definitions.TYPE_INDICATOR_CS, location='/',
        parent=self._gpt_path_spec)
    self.assertTrue(file_system.FileEntryExistsByPathSpec(path_spec))

    path_spec = path_spec_factory.Factory.NewPathSpec(
        definitions.TYPE_INDICATOR_CS, parent=self._gpt_path_spec,
        volume_index=0)
    self.assertTrue(file_system.FileEntryExistsByPathSpec(path_spec))

    path_spec = path_spec_factory.Factory.NewPathSpec(
        definitions.TYPE_INDICATOR_CS, location='/cs1',
        parent=self._gpt_path_spec)
    self.assertTrue(file_system.FileEntryExistsByPathSpec(path_spec))

    path_spec = path_spec_factory.Factory.NewPathSpec(
        definitions.TYPE_INDICATOR_CS, parent=self._gpt_path_spec,
        volume_index=9)
    self.assertFalse(file_system.FileEntryExistsByPathSpec(path_spec))

    path_spec = path_spec_factory.Factory.NewPathSpec(
        definitions.TYPE_INDICATOR_CS, location='/cs0',
        parent=self._gpt_path_spec)
    self.assertFalse(file_system.FileEntryExistsByPathSpec(path_spec))

    path_spec = path_spec_factory.Factory.NewPathSpec(
        definitions.TYPE_INDICATOR_CS, location='/cs9',
        parent=self._gpt_path_spec)
    self.assertFalse(file_system.FileEntryExistsByPathSpec(path_spec))

  def testGetFileEntryByPathSpec(self):
    """Tests the GetFileEntryByPathSpec function."""
    file_system = cs_file_system.CSFileSystem(
        self._resolver_context, self._cs_path_spec)
    self.assertIsNotNone(file_system)

    file_system.Open()

    path_spec = path_spec_factory.Factory.NewPathSpec(
        definitions.TYPE_INDICATOR_CS, location='/',
        parent=self._gpt_path_spec)
    file_entry = file_system.GetFileEntryByPathSpec(path_spec)

    self.assertIsNotNone(file_entry)
    self.assertEqual(file_entry.name, '')

    path_spec = path_spec_factory.Factory.NewPathSpec(
        definitions.TYPE_INDICATOR_CS, parent=self._gpt_path_spec,
        volume_index=0)
    file_entry = file_system.GetFileEntryByPathSpec(path_spec)

    self.assertIsNotNone(file_entry)
    self.assertEqual(file_entry.name, 'cs1')

    path_spec = path_spec_factory.Factory.NewPathSpec(
        definitions.TYPE_INDICATOR_CS, location='/cs1',
        parent=self._gpt_path_spec)
    file_entry = file_system.GetFileEntryByPathSpec(path_spec)

    self.assertIsNotNone(file_entry)
    self.assertEqual(file_entry.name, 'cs1')

    path_spec = path_spec_factory.Factory.NewPathSpec(
        definitions.TYPE_INDICATOR_CS, parent=self._gpt_path_spec,
        volume_index=9)
    file_entry = file_system.GetFileEntryByPathSpec(path_spec)

    self.assertIsNone(file_entry)

    path_spec = path_spec_factory.Factory.NewPathSpec(
        definitions.TYPE_INDICATOR_CS, location='/cs0',
        parent=self._gpt_path_spec)
    file_entry = file_system.GetFileEntryByPathSpec(path_spec)

    self.assertIsNone(file_entry)

    path_spec = path_spec_factory.Factory.NewPathSpec(
        definitions.TYPE_INDICATOR_CS, location='/cs9',
        parent=self._gpt_path_spec)
    file_entry = file_system.GetFileEntryByPathSpec(path_spec)

    self.assertIsNone(file_entry)

  def testGetRootFileEntry(self):
    """Test the get root file entry functionality."""
    file_system = cs_file_system.CSFileSystem(
        self._resolver_context, self._cs_path_spec)
    self.assertIsNotNone(file_system)

    file_system.Open()

    file_entry = file_system.GetRootFileEntry()

    self.assertIsNotNone(file_entry)
    self.assertEqual(file_entry.name, '')


if __name__ == '__main__':
  unittest.main()



# -*- coding: utf-8 -*-
"""The XFS directory implementation."""

from dfvfs.path import xfs_path_spec
from dfvfs.vfs import directory


class XFSDirectory(directory.Directory):
  """File system directory that uses pyfsxfs."""

  def __init__(self, file_system, path_spec, fsxfs_file_entry):
    """Initializes a directory.

    Args:
      file_system (FileSystem): file system.
      path_spec (PathSpec): path specification.
      fsxfs_file_entry (pyfsxfs.file_entry): XFS file entry.
    """
    super(XFSDirectory, self).__init__(file_system, path_spec)
    self._fsxfs_file_entry = fsxfs_file_entry

  def _EntriesGenerator(self):
    """Retrieves directory entries.

    Since a directory can contain a vast number of entries using
    a generator is more memory efficient.

    Yields:
      XFSPathSpec: XFS path specification.
    """
    location = getattr(self.path_spec, 'location', None)

    for fsxfs_sub_file_entry in self._fsxfs_file_entry.sub_file_entries:
      directory_entry = fsxfs_sub_file_entry.name

      if not location or location == self._file_system.PATH_SEPARATOR:
        directory_entry = self._file_system.JoinPath([directory_entry])
      else:
        directory_entry = self._file_system.JoinPath([
            location, directory_entry])

      yield xfs_path_spec.XFSPathSpec(
          inode=fsxfs_sub_file_entry.inode_number, location=directory_entry,
          parent=self.path_spec.parent)



# -*- coding: utf-8 -*-
"""Helper functions for GUID Partition Table (GPT) support."""


def GPTPathSpecGetEntryIndex(path_spec):
  """Retrieves the entry index from the path specification.

  Args:
    path_spec (PathSpec): path specification.

  Returns:
    int: entry index or None if not available.
  """
  entry_index = getattr(path_spec, 'entry_index', None)

  if entry_index is None:
    location = getattr(path_spec, 'location', None)

    if location is None or not location.startswith('/p'):
      return None

    entry_index = None
    try:
      entry_index = int(location[2:], 10) - 1
    except ValueError:
      pass

    if entry_index is None or entry_index < 0:
      return None

  return entry_index



# -*- coding: utf-8 -*-
"""The BitLocker Drive Encryption (BDE) path specification implementation."""

from dfvfs.lib import definitions
from dfvfs.path import factory
from dfvfs.path import path_spec


class BDEPathSpec(path_spec.PathSpec):
  """BDE path specification.

  Attributes:
    password (str): password.
    recovery_password (str): recovery password.
    startup_key (str): name of the startup key file.
  """

  TYPE_INDICATOR = definitions.TYPE_INDICATOR_BDE

  def __init__(
      self, password=None, parent=None, recovery_password=None,
      startup_key=None, **kwargs):
    """Initializes a path specification.

    Note that the BDE path specification must have a parent.

    Args:
      password (Optional[str]): password.
      parent (Optional[PathSpec]): parent path specification.
      recovery_password (Optional[str]): recovery password.
      startup_key (Optional[str]): name of the startup key file.

    Raises:
      ValueError: when parent is not set.
    """
    if not parent:
      raise ValueError('Missing parent value.')

    super(BDEPathSpec, self).__init__(parent=parent, **kwargs)
    self.password = password
    self.recovery_password = recovery_password
    self.startup_key = startup_key

  @property
  def comparable(self):
    """str: comparable representation of the path specification."""
    string_parts = []

    if self.password:
      string_parts.append(f'password: {self.password:s}')
    if self.recovery_password:
      string_parts.append(f'recovery_password: {self.recovery_password:s}')
    if self.startup_key:
      string_parts.append(f'startup_key: {self.startup_key:s}')

    return self._GetComparable(sub_comparable_string=', '.join(string_parts))


factory.Factory.RegisterPathSpec(BDEPathSpec)



# -*- coding: utf-8 -*-
"""The Virtual File System (VFS) path specification interface."""

from dfvfs.lib import definitions


class PathSpec(object):
  """Path specification interface.

  Attributes:
    parent (PathSpec): parent path specification.
  """

  # pylint: disable=missing-raises-doc

  _IS_SYSTEM_LEVEL = False

  def __init__(self, parent=None, **kwargs):
    """Initializes a path specification.

    Args:
      parent (Optional[PathSpec]): parent path specification.
      kwargs (dict[str, object]): keyword arguments depending on the path
          specification.

    Raises:
      ValueError: if a derived path specification class does not define a type
          indicator or when there are unused keyword arguments.
    """
    if kwargs:
      keyword_arguments = ', '.join(kwargs)
      raise ValueError(f'Unused keyword arguments: {keyword_arguments:s}.')

    super(PathSpec, self).__init__()
    self.parent = parent

    if not getattr(self, 'TYPE_INDICATOR', None):
      raise ValueError('Missing type indicator.')

  def __eq__(self, other):
    """Determines if the path specification is equal to the other."""
    return isinstance(other, PathSpec) and self.comparable == other.comparable

  def __hash__(self):
    """Returns the hash of a path specification."""
    return hash(self.comparable)

  def _GetComparable(self, sub_comparable_string=''):
    """Retrieves the comparable representation.

    This is a convenience function for constructing comparables.

    Args:
      sub_comparable_string (str): sub comparable string.

    Returns:
      str: comparable representation of the path specification.
    """
    string_parts = []

    string_parts.append(getattr(self.parent, 'comparable', ''))
    string_parts.append(f'type: {self.type_indicator:s}')

    if sub_comparable_string:
      string_parts.append(f', {sub_comparable_string:s}')
    string_parts.append('\n')

    return ''.join(string_parts)

  @property
  def comparable(self):
    """str: comparable representation of the path specification."""
    return self._GetComparable()

  @property
  def type_indicator(self):
    """str: type indicator."""
    # pylint: disable=no-member
    return self.TYPE_INDICATOR

  def CopyToDict(self):
    """Copies the path specification to a dictionary.

    Returns:
      dict[str, object]: path specification attributes.
    """
    path_spec_dict = {}
    for attribute_name, attribute_value in self.__dict__.items():
      if attribute_value is None:
        continue

      if attribute_name == 'parent':
        attribute_value = attribute_value.CopyToDict()

      path_spec_dict[attribute_name] = attribute_value

    return path_spec_dict

  def HasParent(self):
    """Determines if the path specification has a parent.

    Returns:
      bool: True if the path specification has a parent.
    """
    return self.parent is not None

  def IsFileSystem(self):
    """Determines if the path specification is a file system.

    Returns:
      bool: True if the path specification is a file system.
    """
    return self.type_indicator in definitions.FILE_SYSTEM_TYPE_INDICATORS

  def IsSystemLevel(self):
    """Determines if the path specification is at system-level.

    System-level is an indication used if the path specification is
    handled by the operating system and should not have a parent.

    Returns:
      bool: True if the path specification is at system-level.
    """
    return getattr(self, '_IS_SYSTEM_LEVEL', False)

  def IsVolumeSystem(self):
    """Determines if the path specification is a volume system.

    Returns:
      bool: True if the path specification is a volume system.
    """
    return self.type_indicator in definitions.VOLUME_SYSTEM_TYPE_INDICATORS

  def IsVolumeSystemRoot(self):
    """Determines if the path specification is the root of a volume system.

    Returns:
      bool: True if the path specification is the root of a volume system.
    """
    return self.IsVolumeSystem() and getattr(self, 'location', None) == '/'



# -*- coding: utf-8 -*-
"""Helpers for command line tools."""

import abc
import codecs
import getpass
import locale
import logging
import sys
import textwrap

try:
  import win32console
except ImportError:
  win32console = None

from dfdatetime import filetime as dfdatetime_filetime

from dfvfs.helpers import volume_scanner
from dfvfs.lib import definitions
from dfvfs.lib import errors


class CLIInputReader(object):
  """Command line interface input reader interface."""

  def __init__(self, encoding='utf-8'):
    """Initializes an input reader.

    Args:
      encoding (Optional[str]): input encoding.
    """
    super(CLIInputReader, self).__init__()
    self._encoding = encoding

  # pylint: disable=redundant-returns-doc
  @abc.abstractmethod
  def Read(self):
    """Reads a string from the input.

    Returns:
      str: input.
    """


class FileObjectInputReader(CLIInputReader):
  """File object command line interface input reader.

  This input reader relies on the file-like object having a readline method.
  """

  def __init__(self, file_object, encoding='utf-8'):
    """Initializes a file object input reader.

    Args:
      file_object (file): file-like object to read from.
      encoding (Optional[str]): input encoding.
    """
    super(FileObjectInputReader, self).__init__(encoding=encoding)
    self._errors = 'strict'
    self._file_object = file_object

  def Read(self):
    """Reads a string from the input.

    Returns:
      str: input.
    """
    input_string = self._file_object.readline()

    if isinstance(input_string, bytes):
      try:
        input_string = codecs.decode(input_string, self._encoding, self._errors)
      except UnicodeDecodeError:
        if self._errors == 'strict':
          logging.error(
              'Unable to properly read input due to encoding error. '
              'Switching to error tolerant encoding which can result in '
              'non Basic Latin (C0) characters to be replaced with "?" or '
              '"\\ufffd".')
          self._errors = 'replace'

        input_string = codecs.decode(input_string, self._encoding, self._errors)

    return input_string


class StdinInputReader(FileObjectInputReader):
  """Stdin command line interface input reader."""

  def __init__(self, encoding='utf-8'):
    """Initializes a stdin input reader.

    Args:
      encoding (Optional[str]): input encoding.
    """
    super(StdinInputReader, self).__init__(sys.stdin, encoding=encoding)


class CLIOutputWriter(object):
  """Command line interface output writer interface."""

  def __init__(self, encoding='utf-8'):
    """Initializes an output writer.

    Args:
      encoding (Optional[str]): output encoding.
    """
    super(CLIOutputWriter, self).__init__()
    self._encoding = encoding

  def Flush(self):
    """Flushes buffered data to the output."""
    return

  @abc.abstractmethod
  def Write(self, string):
    """Writes a string to the output.

    Args:
      string (str): output.
    """


class FileObjectOutputWriter(CLIOutputWriter):
  """File object command line interface output writer.

  This output writer relies on the file-like object having a write method.
  """

  def __init__(self, file_object, encoding='utf-8'):
    """Initializes a file object output writer.

    Args:
      file_object (file): file-like object to read from.
      encoding (Optional[str]): output encoding.
    """
    super(FileObjectOutputWriter, self).__init__(encoding=encoding)
    self._errors = 'strict'
    self._file_object = file_object

  def Write(self, string):
    """Writes a string to the output.

    Args:
      string (str): output.
    """
    try:
      # Note that encode() will first convert string into a Unicode string
      # if necessary.
      encoded_string = codecs.encode(string, self._encoding, self._errors)
    except UnicodeEncodeError:
      if self._errors == 'strict':
        logging.error(
            'Unable to properly write output due to encoding error. '
            'Switching to error tolerant encoding which can result in '
            'non Basic Latin (C0) characters to be replaced with "?" or '
            '"\\ufffd".')
        self._errors = 'replace'

      encoded_string = codecs.encode(string, self._encoding, self._errors)

    self._file_object.write(encoded_string)


class StdoutOutputWriter(CLIOutputWriter):
  """Stdout command line interface output writer."""

  def __init__(self, encoding='utf-8'):
    """Initializes a stdout output writer.

    Args:
      encoding (Optional[str]): output encoding.
    """
    super(StdoutOutputWriter, self).__init__(encoding=encoding)

  def Flush(self):
    """Flushes buffered data to the output."""
    sys.stdout.flush()

  def Write(self, string):
    """Writes a string to the output.

    Args:
      string (str): output.
    """
    sys.stdout.write(string)


class CLITabularTableView(object):
  """Command line interface tabular table view."""

  _NUMBER_OF_SPACES_IN_TAB = 8

  def __init__(self, column_names=None, column_sizes=None):
    """Initializes a command line interface tabular table view.

    Args:
      column_names (Optional[list[str]]): column names.
      column_sizes (Optional[list[int]]): minimum column sizes, in number of
          characters. If a column name or row value is larger than the
          minimum column size the column will be enlarged. Note that the
          minimum columns size will be rounded up to the number of spaces
          of the next tab.
    """
    super(CLITabularTableView, self).__init__()
    self._columns = column_names or []
    self._column_sizes = column_sizes or []
    self._number_of_columns = len(self._columns)
    self._rows = []

  def _WriteRow(self, output_writer, values, in_bold=False):
    """Writes a row of values aligned with the width to the output writer.

    Args:
      output_writer (CLIOutputWriter): output writer.
      values (list[object]): values.
      in_bold (Optional[bool]): True if the row should be written in bold.
    """
    row_strings = []
    for value_index, value_string in enumerate(values):
      padding_size = self._column_sizes[value_index] - len(value_string)
      padding_string = ' ' * padding_size

      row_strings.extend([value_string, padding_string])

    row_strings.pop()

    row_strings = ''.join(row_strings)

    if in_bold and not win32console:
      # TODO: for win32console get current color and set intensity,
      # write the header separately then reset intensity.
      output_writer.Write(f'\x1b[1m{row_strings:s}\x1b[0m\n')
    else:
      output_writer.Write(f'{row_strings:s}\n')

  def AddRow(self, values):
    """Adds a row of values.

    Args:
      values (list[object]): values.

    Raises:
      ValueError: if the number of values is out of bounds.
    """
    if self._number_of_columns and len(values) != self._number_of_columns:
      raise ValueError('Number of values is out of bounds.')

    if not self._column_sizes and self._columns:
      self._column_sizes = [len(column) for column in self._columns]

    value_strings = []
    for value_index, value_string in enumerate(values):
      if not isinstance(value_string, str):
        value_string = f'{value_string!s}'
      value_strings.append(value_string)

      self._column_sizes[value_index] = max(
          self._column_sizes[value_index], len(value_string))

    self._rows.append(value_strings)

    if not self._number_of_columns:
      self._number_of_columns = len(value_strings)

  def Write(self, output_writer):
    """Writes the table to output writer.

    Args:
      output_writer (CLIOutputWriter): output writer.
    """
    # Round up the column sizes to the nearest tab.
    for column_index, column_size in enumerate(self._column_sizes):
      column_size, _ = divmod(column_size, self._NUMBER_OF_SPACES_IN_TAB)
      column_size = (column_size + 1) * self._NUMBER_OF_SPACES_IN_TAB
      self._column_sizes[column_index] = column_size

    if self._columns:
      self._WriteRow(output_writer, self._columns, in_bold=True)

    for values in self._rows:
      self._WriteRow(output_writer, values)


class CLIVolumeScannerMediator(volume_scanner.VolumeScannerMediator):
  """Command line volume scanner mediator."""

  # For context see: http://en.wikipedia.org/wiki/Byte
  _UNITS_1000 = ['B', 'kB', 'MB', 'GB', 'TB', 'EB', 'ZB', 'YB']
  _UNITS_1024 = ['B', 'KiB', 'MiB', 'GiB', 'TiB', 'EiB', 'ZiB', 'YiB']

  _USER_PROMPT_APFS = (
      'Please specify the identifier(s) of the APFS volume that should be '
      'processed: Note that a range of volumes can be defined as: 3..5. '
      'Multiple volumes can be defined as: 1,3,5 (a list of comma separated '
      'values). Ranges and lists can also be combined as: 1,3..5. The first '
      'volume is 1. All volumes can be defined as "all". If no volumes are '
      'specified none will be processed. You can abort with Ctrl^C.')

  _USER_PROMPT_LVM = (
      'Please specify the identifier(s) of the LVM volume that should be '
      'processed: Note that a range of volumes can be defined as: 3..5. '
      'Multiple volumes can be defined as: 1,3,5 (a list of comma separated '
      'values). Ranges and lists can also be combined as: 1,3..5. The first '
      'volume is 1. All volumes can be defined as "all". If no volumes are '
      'specified none will be processed. You can abort with Ctrl^C.')

  _USER_PROMPT_PARTITIONS = (
      'Please specify the identifier of the partition that should be '
      'processed. All partitions can be defined as: "all". Note that you can '
      'abort with Ctrl^C.')

  _USER_PROMPT_VSS = (
      'Please specify the identifier(s) of the VSS that should be processed: '
      'Note that a range of stores can be defined as: 3..5. Multiple stores '
      'can be defined as: 1,3,5 (a list of comma separated values). Ranges '
      'and lists can also be combined as: 1,3..5. The first store is 1. All '
      'stores can be defined as "all". If no stores are specified none will '
      'be processed. You can abort with Ctrl^C.')

  def __init__(self, input_reader=None, output_writer=None):
    """Initializes a volume scanner mediator.

    Args:
      input_reader (Optional[CLIInputReader]): input reader, where None
          indicates that the stdin input reader should be used.
      output_writer (Optional[CLIOutputWriter]): output writer, where None
          indicates that the stdout output writer should be used.
    """
    preferred_encoding = locale.getpreferredencoding()
    if not input_reader:
      input_reader = StdinInputReader(encoding=preferred_encoding)
    if not output_writer:
      output_writer = StdoutOutputWriter(encoding=preferred_encoding)

    super(CLIVolumeScannerMediator, self).__init__()
    self._input_reader = input_reader
    self._output_writer = output_writer
    self._textwrapper = textwrap.TextWrapper()

  def _FormatHumanReadableSize(self, size):
    """Represents a number of bytes as a human readable string.

    Args:
      size (int): size in bytes.

    Returns:
      str: human readable string of the size.
    """
    magnitude_1000 = 0
    size_1000 = float(size)
    while size_1000 >= 1000:
      size_1000 /= 1000
      magnitude_1000 += 1

    magnitude_1024 = 0
    size_1024 = float(size)
    while size_1024 >= 1024:
      size_1024 /= 1024
      magnitude_1024 += 1

    size_string_1000 = None
    if 0 < magnitude_1000 <= 7:
      units_1000 = self._UNITS_1000[magnitude_1000]
      size_string_1000 = f'{size_1000:.1f}{units_1000:s}'

    size_string_1024 = None
    if 0 < magnitude_1024 <= 7:
      units_1024 = self._UNITS_1024[magnitude_1024]
      size_string_1024 = f'{size_1024:.1f}{units_1024:s}'

    if not size_string_1000 or not size_string_1024:
      return f'{size:d} B'

    return f'{size_string_1024:s} / {size_string_1000:s} ({size:d} B)'

  def _ParseVolumeIdentifiersString(
      self, volume_identifiers_string, prefix='v'):
    """Parses a user specified volume identifiers string.

    Args:
      volume_identifiers_string (str): user specified volume identifiers. A
          range of volumes can be defined as: "3..5". Multiple volumes can be
          defined as: "1,3,5" (a list of comma separated values). Ranges and
          lists can also be combined as: "1,3..5". The first volume is 1. All
          volumes can be defined as "all". No volumes can be defined as an
          empty string or "none".
      prefix (Optional[str]): volume identifier prefix.

    Returns:
      list[str]: volume identifiers with prefix or the string "all".

    Raises:
      ValueError: if the volume identifiers string is invalid.
    """
    prefix_length = 0
    if prefix:
      prefix_length = len(prefix)

    if not volume_identifiers_string or volume_identifiers_string == 'none':
      return []

    if volume_identifiers_string == 'all':
      return ['all']

    volume_identifiers = set()
    for identifiers_range in volume_identifiers_string.split(','):
      # Determine if the range is formatted as 1..3 otherwise it indicates
      # a single volume identifier.
      if '..' in identifiers_range:
        first_identifier, last_identifier = identifiers_range.split('..')

        if prefix and first_identifier.startswith(prefix):
          first_identifier = first_identifier[prefix_length:]

        if prefix and last_identifier.startswith(prefix):
          last_identifier = last_identifier[prefix_length:]

        try:
          first_identifier = int(first_identifier, 10)
          last_identifier = int(last_identifier, 10)
        except ValueError:
          raise ValueError(
              f'Invalid volume identifiers range: {identifiers_range:s}.')

        for volume_identifier in range(first_identifier, last_identifier + 1):
          if volume_identifier not in volume_identifiers:
            if prefix:
              volume_identifier = f'{prefix:s}{volume_identifier:d}'
            volume_identifiers.add(volume_identifier)
      else:
        identifier = identifiers_range
        if prefix and identifier.startswith(prefix):
          identifier = identifiers_range[prefix_length:]

        try:
          volume_identifier = int(identifier, 10)
        except ValueError:
          raise ValueError(
              f'Invalid volume identifier range: {identifiers_range:s}.')

        if prefix:
          volume_identifier = f'{prefix:s}{volume_identifier:d}'
        volume_identifiers.add(volume_identifier)

    # Note that sorted will return a list.
    return sorted(volume_identifiers)

  def _PrintAPFSVolumeIdentifiersOverview(
      self, volume_system, volume_identifiers):
    """Prints an overview of APFS volume identifiers.

    Args:
      volume_system (APFSVolumeSystem): volume system.
      volume_identifiers (list[str]): allowed volume identifiers.

    Raises:
      ScannerError: if a volume cannot be resolved from the volume identifier.
    """
    header = 'The following Apple File System (APFS) volumes were found:\n'
    self._output_writer.Write(header)

    column_names = ['Identifier', 'Name']
    table_view = CLITabularTableView(column_names=column_names)

    # Sort the volume identifiers in alphanumeric order.
    for volume_identifier in sorted(volume_identifiers, key=lambda string: int(
        ''.join([character for character in string if character.isdigit()]))):
      volume = volume_system.GetVolumeByIdentifier(volume_identifier)
      if not volume:
        raise errors.ScannerError(
            f'Volume missing for identifier: {volume_identifier:s}.')

      volume_attribute = volume.GetAttribute('name')
      table_view.AddRow([volume.identifier, volume_attribute.value])

    self._output_writer.Write('\n')
    table_view.Write(self._output_writer)

  def _PrintLVMVolumeIdentifiersOverview(
      self, volume_system, volume_identifiers):
    """Prints an overview of LVM volume identifiers.

    Args:
      volume_system (LVMVolumeSystem): volume system.
      volume_identifiers (list[str]): allowed volume identifiers.

    Raises:
      ScannerError: if a volume cannot be resolved from the volume identifier.
    """
    header = 'The following Logical Volume Manager (LVM) volumes were found:\n'
    self._output_writer.Write(header)

    column_names = ['Identifier']
    table_view = CLITabularTableView(column_names=column_names)

    # Sort the volume identifiers in alphanumeric order.
    for volume_identifier in sorted(volume_identifiers, key=lambda string: int(
        ''.join([character for character in string if character.isdigit()]))):
      volume = volume_system.GetVolumeByIdentifier(volume_identifier)
      if not volume:
        raise errors.ScannerError(
            f'Volume missing for identifier: {volume_identifier:s}.')

      table_view.AddRow([volume.identifier])

    self._output_writer.Write('\n')
    table_view.Write(self._output_writer)

  def _PrintPartitionIdentifiersOverview(
      self, volume_system, volume_identifiers):
    """Prints an overview of partition identifiers.

    Args:
      volume_system (VolumeSystem): volume system.
      volume_identifiers (list[str]): allowed volume identifiers.

    Raises:
      ScannerError: if a volume cannot be resolved from the volume identifier.
    """
    header = 'The following partitions were found:\n'
    self._output_writer.Write(header)

    column_names = ['Identifier', 'Offset (in bytes)', 'Size (in bytes)']
    table_view = CLITabularTableView(column_names=column_names)

    # Sort the volume identifiers in alphanumeric order.
    for volume_identifier in sorted(volume_identifiers, key=lambda string: int(
        ''.join([character for character in string if character.isdigit()]))):
      volume = volume_system.GetVolumeByIdentifier(volume_identifier)
      if not volume:
        raise errors.ScannerError(
            f'Partition missing for identifier: {volume_identifier:s}.')

      volume_extent = volume.extents[0]
      volume_offset = f'{volume_extent.offset:d} (0x{volume_extent.offset:08x})'
      volume_size = self._FormatHumanReadableSize(volume_extent.size)

      table_view.AddRow([volume.identifier, volume_offset, volume_size])

    self._output_writer.Write('\n')
    table_view.Write(self._output_writer)

  def _PrintVSSStoreIdentifiersOverview(
      self, volume_system, volume_identifiers):
    """Prints an overview of VSS store identifiers.

    Args:
      volume_system (VShadowVolumeSystem): volume system.
      volume_identifiers (list[str]): allowed volume identifiers.

    Raises:
      ScannerError: if a volume cannot be resolved from the volume identifier.
    """
    header = 'The following Volume Shadow Snapshots (VSS) were found:\n'
    self._output_writer.Write(header)

    column_names = ['Identifier', 'Creation Time']
    table_view = CLITabularTableView(column_names=column_names)

    # Sort the volume identifiers in alphanumeric order.
    for volume_identifier in sorted(volume_identifiers, key=lambda string: int(
        ''.join([character for character in string if character.isdigit()]))):
      volume = volume_system.GetVolumeByIdentifier(volume_identifier)
      if not volume:
        raise errors.ScannerError(
            f'Volume missing for identifier: {volume_identifier:s}.')

      volume_attribute = volume.GetAttribute('creation_time')
      filetime = dfdatetime_filetime.Filetime(timestamp=volume_attribute.value)
      creation_time = filetime.CopyToDateTimeString()

      if volume.HasExternalData():
        creation_time = ''.join([
            creation_time, '\tWARNING: data stored outside volume'])

      table_view.AddRow([volume.identifier, creation_time])

    self._output_writer.Write('\n')
    table_view.Write(self._output_writer)

  def _ReadSelectedVolumes(self, volume_system, prefix='v'):
    """Reads the selected volumes provided by the user.

    Args:
      volume_system (VolumeSystem): volume system.
      prefix (Optional[str]): volume identifier prefix.

    Returns:
      list[str]: selected volume identifiers including prefix.

    Raises:
      KeyboardInterrupt: if the user requested to abort.
      ValueError: if the volume identifiers string could not be parsed.
    """
    volume_identifiers_string = self._input_reader.Read()
    volume_identifiers_string = volume_identifiers_string.strip()

    if not volume_identifiers_string:
      return []

    selected_volumes = self._ParseVolumeIdentifiersString(
        volume_identifiers_string, prefix=prefix)

    if selected_volumes == ['all']:
      return volume_system.volume_identifiers

    return selected_volumes

  def GetAPFSVolumeIdentifiers(self, volume_system, volume_identifiers):
    """Retrieves APFS volume identifiers.

    This method can be used to prompt the user to provide APFS volume
    identifiers.

    Args:
      volume_system (APFSVolumeSystem): volume system.
      volume_identifiers (list[str]): volume identifiers including prefix.

    Returns:
      list[str]: selected volume identifiers including prefix or None.
    """
    self._PrintAPFSVolumeIdentifiersOverview(volume_system, volume_identifiers)

    while True:
      self._output_writer.Write('\n')

      lines = self._textwrapper.wrap(self._USER_PROMPT_APFS)
      self._output_writer.Write('\n'.join(lines))
      self._output_writer.Write('\n\nVolume identifier(s): ')

      try:
        selected_volumes = self._ReadSelectedVolumes(
            volume_system, prefix='apfs')
        if (not selected_volumes or
            not set(selected_volumes).difference(volume_identifiers)):
          break
      except ValueError:
        pass

      self._output_writer.Write('\n')

      lines = self._textwrapper.wrap(
          'Unsupported volume identifier(s), please try again or abort with '
          'Ctrl^C.')
      self._output_writer.Write('\n'.join(lines))
      self._output_writer.Write('\n\n')

    return selected_volumes

  def GetLVMVolumeIdentifiers(self, volume_system, volume_identifiers):
    """Retrieves LVM volume identifiers.

    This method can be used to prompt the user to provide LVM volume
    identifiers.

    Args:
      volume_system (LVMVolumeSystem): volume system.
      volume_identifiers (list[str]): volume identifiers including prefix.

    Returns:
      list[str]: selected volume identifiers including prefix or None.
    """
    self._PrintLVMVolumeIdentifiersOverview(volume_system, volume_identifiers)

    while True:
      self._output_writer.Write('\n')

      lines = self._textwrapper.wrap(self._USER_PROMPT_LVM)
      self._output_writer.Write('\n'.join(lines))
      self._output_writer.Write('\n\nVolume identifier(s): ')

      try:
        selected_volumes = self._ReadSelectedVolumes(
            volume_system, prefix='lvm')
        if (not selected_volumes or
            not set(selected_volumes).difference(volume_identifiers)):
          break
      except ValueError:
        pass

      self._output_writer.Write('\n')

      lines = self._textwrapper.wrap(
          'Unsupported volume identifier(s), please try again or abort with '
          'Ctrl^C.')
      self._output_writer.Write('\n'.join(lines))
      self._output_writer.Write('\n\n')

    return selected_volumes

  def GetPartitionIdentifiers(self, volume_system, volume_identifiers):
    """Retrieves partition identifiers.

    This method can be used to prompt the user to provide partition identifiers.

    Args:
      volume_system (VolumeSystem): volume system.
      volume_identifiers (list[str]): volume identifiers including prefix.

    Returns:
      list[str]: selected volume identifiers including prefix or None.
    """
    self._PrintPartitionIdentifiersOverview(volume_system, volume_identifiers)

    while True:
      self._output_writer.Write('\n')

      lines = self._textwrapper.wrap(self._USER_PROMPT_PARTITIONS)
      self._output_writer.Write('\n'.join(lines))
      self._output_writer.Write('\n\nPartition identifier(s): ')

      try:
        selected_volumes = self._ReadSelectedVolumes(volume_system, prefix='p')
        if (not selected_volumes or
            not set(selected_volumes).difference(volume_identifiers)):
          break
      except ValueError:
        pass

      self._output_writer.Write('\n')

      lines = self._textwrapper.wrap(
          'Unsupported partition identifier(s), please try again or abort with '
          'Ctrl^C.')
      self._output_writer.Write('\n'.join(lines))
      self._output_writer.Write('\n\n')

    return selected_volumes

  def GetVSSStoreIdentifiers(self, volume_system, volume_identifiers):
    """Retrieves VSS store identifiers.

    This method can be used to prompt the user to provide VSS store identifiers.

    Args:
      volume_system (VShadowVolumeSystem): volume system.
      volume_identifiers (list[str]): volume identifiers including prefix.

    Returns:
      list[str]: selected volume identifiers including prefix or None.
    """
    self._PrintVSSStoreIdentifiersOverview(volume_system, volume_identifiers)

    while True:
      self._output_writer.Write('\n')

      lines = self._textwrapper.wrap(self._USER_PROMPT_VSS)
      self._output_writer.Write('\n'.join(lines))
      self._output_writer.Write('\n\nVSS identifier(s): ')

      try:
        selected_volumes = self._ReadSelectedVolumes(
            volume_system, prefix='vss')
        if (not selected_volumes or
            not set(selected_volumes).difference(volume_identifiers)):
          break
      except ValueError:
        pass

      self._output_writer.Write('\n')

      lines = self._textwrapper.wrap(
          'Unsupported VSS identifier(s), please try again or abort with '
          'Ctrl^C.')
      self._output_writer.Write('\n'.join(lines))
      self._output_writer.Write('\n\n')

    return selected_volumes

  def ParseVolumeIdentifiersString(self, volume_identifiers_string):
    """Parses a user specified volume identifiers string.

    Args:
      volume_identifiers_string (str): user specified volume identifiers. A
          range of volumes can be defined as: "3..5". Multiple volumes can be
          defined as: "1,3,5" (a list of comma separated values). Ranges and
          lists can also be combined as: "1,3..5". The first volume is 1. All
          volumes can be defined as "all". No volumes can be defined as an
          empty string or "none".

    Returns:
      list[str]: volume identifiers with prefix or the string "all".

    Raises:
      ValueError: if the volume identifiers string is invalid.
    """
    return self._ParseVolumeIdentifiersString(
        volume_identifiers_string, prefix=None)

  def PrintWarning(self, warning):
    """Prints a warning.

    Args:
      warning (str): warning text.
    """
    self._output_writer.Write(f'[WARNING] {warning:s}\n\n')

  def UnlockEncryptedVolume(
      self, source_scanner_object, scan_context, locked_scan_node, credentials):
    """Unlocks an encrypted volume.

    This method can be used to prompt the user to provide encrypted volume
    credentials.

    Args:
      source_scanner_object (SourceScanner): source scanner.
      scan_context (SourceScannerContext): source scanner context.
      locked_scan_node (SourceScanNode): locked scan node.
      credentials (Credentials): credentials supported by the locked scan node.

    Returns:
      bool: True if the volume was unlocked.
    """
    # TODO: print volume description.
    if locked_scan_node.type_indicator == (
        definitions.TYPE_INDICATOR_APFS_CONTAINER):
      header = 'Found an APFS encrypted volume.\n\n'
    elif locked_scan_node.type_indicator == definitions.TYPE_INDICATOR_BDE:
      header = 'Found a BitLocker encrypted volume.\n\n'
    elif locked_scan_node.type_indicator == definitions.TYPE_INDICATOR_CS:
      header = 'Found a CoreStorage (FileVault) encrypted volume.\n\n'
    elif locked_scan_node.type_indicator == definitions.TYPE_INDICATOR_LUKSDE:
      header = 'Found a LUKS encrypted volume.\n\n'
    else:
      header = 'Found an encrypted volume.\n\n'

    self._output_writer.Write(header)

    credentials_list = sorted(credentials.CREDENTIALS)
    credentials_list.append('skip')

    self._output_writer.Write('Supported credentials:\n\n')

    for index, name in enumerate(credentials_list):
      credential_index = index + 1
      self._output_writer.Write(f'  {credential_index:d}. {name:s}\n')

    self._output_writer.Write('\nNote that you can abort with Ctrl^C.\n\n')

    result = False
    while not result:
      self._output_writer.Write('Select a credential to unlock the volume: ')
      self._output_writer.Flush()

      input_line = self._input_reader.Read()
      input_line = input_line.strip()

      if input_line in credentials_list:
        credential_type = input_line
      else:
        try:
          credential_type = int(input_line, 10)
          credential_type = credentials_list[credential_type - 1]
        except (IndexError, ValueError):
          self._output_writer.Write(f'Unsupported credential: {input_line:s}\n')
          continue

      if credential_type == 'skip':
        break

      credential_data = getpass.getpass('Enter credential data: ')
      self._output_writer.Write('\n')

      if credential_type == 'key':
        try:
          credential_data = codecs.decode(credential_data, 'hex')
        except TypeError:
          self._output_writer.Write('Unsupported credential data.\n')
          continue

      try:
        result = source_scanner_object.Unlock(
            scan_context, locked_scan_node.path_spec, credential_type,
            credential_data)
      except errors.BackEndError:
        result = False

      if not result:
        self._output_writer.Write('Unable to unlock volume.\n\n')

    return result



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Tests for the resolver context object."""

import platform
import unittest

from dfvfs.file_io import fake_file_io
from dfvfs.lib import definitions
from dfvfs.lib import errors
from dfvfs.path import factory as path_spec_factory
from dfvfs.path import fake_path_spec
from dfvfs.resolver import context
from dfvfs.resolver import resolver
from dfvfs.vfs import fake_file_system

from tests import test_lib as shared_test_lib


class ContextTest(shared_test_lib.BaseTestCase):
  """Tests for the resolver context object."""

  # pylint: disable=protected-access

  def testCacheFileObject(self):
    """Tests the cache file-like object functionality."""
    resolver_context = context.Context()

    self.assertEqual(len(resolver_context._file_object_cache), 0)

    path_spec = fake_path_spec.FakePathSpec(location='/empty.txt')
    file_object = fake_file_io.FakeFile(resolver_context, path_spec, b'')

    resolver_context.CacheFileObject(path_spec, file_object)
    self.assertEqual(len(resolver_context._file_object_cache), 1)

    cached_object = resolver_context.GetFileObject(path_spec)
    self.assertEqual(cached_object, file_object)

  def testCacheFileSystem(self):
    """Tests the cache file system object functionality."""
    resolver_context = context.Context()

    self.assertEqual(len(resolver_context._file_system_cache), 0)

    path_spec = fake_path_spec.FakePathSpec(location='/')
    file_system = fake_file_system.FakeFileSystem(resolver_context, path_spec)

    resolver_context.CacheFileSystem(path_spec, file_system)
    self.assertEqual(len(resolver_context._file_system_cache), 1)

    cached_object = resolver_context.GetFileSystem(path_spec)
    self.assertEqual(cached_object, file_system)

  def testGetMountPoint(self):
    """Tests the GetMountPoint function."""
    test_path = self._GetTestFilePath(['ext2.qcow2'])
    self._SkipIfPathNotExists(test_path)

    test_os_path_spec = path_spec_factory.Factory.NewPathSpec(
        definitions.TYPE_INDICATOR_OS, location=test_path)
    test_qcow_path_spec = path_spec_factory.Factory.NewPathSpec(
        definitions.TYPE_INDICATOR_QCOW, parent=test_os_path_spec)

    resolver_context = context.Context()
    resolver_context.RegisterMountPoint('C', test_qcow_path_spec)

    try:
      mount_point_path_spec = resolver_context.GetMountPoint('C')
      self.assertEqual(mount_point_path_spec, test_qcow_path_spec)

      mount_point_path_spec = resolver_context.GetMountPoint('D')
      self.assertIsNone(mount_point_path_spec)

    finally:
      resolver_context.DeregisterMountPoint('C')

  def testOpenFileObjectOnDirectory(self):
    """Tests mount point resolving on a directory."""
    test_path = self._GetTestFilePath(['testdir_os'])
    self._SkipIfPathNotExists(test_path)

    test_os_path_spec = path_spec_factory.Factory.NewPathSpec(
        definitions.TYPE_INDICATOR_OS, location=test_path)

    resolver_context = context.Context()
    resolver_context.RegisterMountPoint('testdir_os', test_os_path_spec)

    if platform.system() == 'Windows':
      test_mounted_location = '\\file1.txt'
    else:
      test_mounted_location = '/file1.txt'

    try:
      parent_path_spec = path_spec_factory.Factory.NewPathSpec(
          definitions.TYPE_INDICATOR_MOUNT, identifier='testdir_os')
      path_spec = path_spec_factory.Factory.NewPathSpec(
          definitions.TYPE_INDICATOR_OS, location=test_mounted_location,
          parent=parent_path_spec)
      file_object = resolver.Resolver.OpenFileObject(
          path_spec, resolver_context=resolver_context)

      self.assertIsNotNone(file_object)
      self.assertEqual(file_object.get_size(), 6)

    finally:
      resolver_context.DeregisterMountPoint('testdir_os')

  def testOpenFileObjectOnImage(self):
    """Tests mount point resolving on a storage media image."""
    test_path = self._GetTestFilePath(['ext2.qcow2'])
    self._SkipIfPathNotExists(test_path)

    test_os_path_spec = path_spec_factory.Factory.NewPathSpec(
        definitions.TYPE_INDICATOR_OS, location=test_path)
    test_qcow_path_spec = path_spec_factory.Factory.NewPathSpec(
        definitions.TYPE_INDICATOR_QCOW, parent=test_os_path_spec)

    resolver_context = context.Context()
    resolver_context.RegisterMountPoint('C', test_qcow_path_spec)

    try:
      parent_path_spec = path_spec_factory.Factory.NewPathSpec(
          definitions.TYPE_INDICATOR_MOUNT, identifier='C')
      path_spec = path_spec_factory.Factory.NewPathSpec(
          definitions.PREFERRED_EXT_BACK_END, location='/passwords.txt',
          parent=parent_path_spec)
      file_object = resolver.Resolver.OpenFileObject(
          path_spec, resolver_context=resolver_context)

      self.assertIsNotNone(file_object)
      self.assertEqual(file_object.get_size(), 116)

      parent_path_spec = path_spec_factory.Factory.NewPathSpec(
          definitions.TYPE_INDICATOR_MOUNT, identifier='D')
      path_spec = path_spec_factory.Factory.NewPathSpec(
          definitions.PREFERRED_EXT_BACK_END, location='/passwords.txt',
          parent=parent_path_spec)

      with self.assertRaises(errors.MountPointError):
        file_object = resolver.Resolver.OpenFileObject(
            path_spec, resolver_context=resolver_context)

    finally:
      resolver_context.DeregisterMountPoint('C')


if __name__ == '__main__':
  unittest.main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Tests for the file-like object implementation using pysmraw."""

import unittest

from dfvfs.lib import definitions
from dfvfs.lib import errors
from dfvfs.path import factory as path_spec_factory

from tests.file_io import test_lib


class RawFileTest(test_lib.Ext2ImageFileTestCase):
  """Tests the RAW storage media image file-like object on a RAW image."""

  def setUp(self):
    """Sets up the needed objects used throughout the test."""
    super(RawFileTest, self).setUp()
    test_path = self._GetTestFilePath(['ext2.raw'])
    self._SkipIfPathNotExists(test_path)

    self._os_path_spec = path_spec_factory.Factory.NewPathSpec(
        definitions.TYPE_INDICATOR_OS, location=test_path)
    self._raw_path_spec = path_spec_factory.Factory.NewPathSpec(
        definitions.TYPE_INDICATOR_RAW, parent=self._os_path_spec)

  def testOpenCloseInode(self):
    """Test the open and close functionality using an inode."""
    self._TestOpenCloseInode(self._raw_path_spec)

  def testOpenCloseLocation(self):
    """Test the open and close functionality using a location."""
    self._TestOpenCloseLocation(self._raw_path_spec)

    # Try open with a path specification that has no parent.
    path_spec = path_spec_factory.Factory.NewPathSpec(
        definitions.TYPE_INDICATOR_RAW, parent=self._os_path_spec)
    path_spec.parent = None

    with self.assertRaises(errors.PathSpecError):
      self._TestOpenCloseLocation(path_spec)

  def testSeek(self):
    """Test the seek functionality."""
    self._TestSeek(self._raw_path_spec)

  def testRead(self):
    """Test the read functionality."""
    self._TestRead(self._raw_path_spec)


class SplitRawFileTest(test_lib.Ext2ImageFileTestCase):
  """Tests the RAW storage media image file-like object on a split RAW image."""

  def setUp(self):
    """Sets up the needed objects used throughout the test."""
    super(SplitRawFileTest, self).setUp()
    test_path = self._GetTestFilePath(['ext2.splitraw.000'])
    self._SkipIfPathNotExists(test_path)

    self._os_path_spec = path_spec_factory.Factory.NewPathSpec(
        definitions.TYPE_INDICATOR_OS, location=test_path)
    self._raw_path_spec = path_spec_factory.Factory.NewPathSpec(
        definitions.TYPE_INDICATOR_RAW, parent=self._os_path_spec)

  def testOpenCloseInode(self):
    """Test the open and close functionality using an inode."""
    self._TestOpenCloseInode(self._raw_path_spec)

  def testOpenCloseLocation(self):
    """Test the open and close functionality using a location."""
    self._TestOpenCloseLocation(self._raw_path_spec)

  def testSeek(self):
    """Test the seek functionality."""
    self._TestSeek(self._raw_path_spec)

  def testRead(self):
    """Test the read functionality."""
    self._TestRead(self._raw_path_spec)


if __name__ == '__main__':
  unittest.main()



# -*- coding: utf-8 -*-
"""The APFS container file entry implementation."""

from dfvfs.lib import apfs_helper
from dfvfs.lib import definitions
from dfvfs.lib import errors
from dfvfs.resolver import resolver
from dfvfs.vfs import apfs_container_directory
from dfvfs.vfs import file_entry


class APFSContainerFileEntry(file_entry.FileEntry):
  """File system file entry that uses pyfsapfs."""

  TYPE_INDICATOR = definitions.TYPE_INDICATOR_APFS_CONTAINER

  def __init__(
      self, resolver_context, file_system, path_spec, is_root=False,
      is_virtual=False):
    """Initializes a file entry.

    Args:
      resolver_context (Context): resolver context.
      file_system (FileSystem): file system.
      path_spec (PathSpec): path specification.
      is_root (Optional[bool]): True if the file entry is the root file entry
          of the corresponding file system.
      is_virtual (Optional[bool]): True if the file entry is a virtual file
          entry emulated by the corresponding file system.

    Raises:
      BackEndError: when the fsapfs volume is missing in a non-virtual
          file entry.
    """
    fsapfs_volume = file_system.GetAPFSVolumeByPathSpec(path_spec)
    if not is_virtual and fsapfs_volume is None:
      raise errors.BackEndError(
          'Missing fsapfs volume in non-virtual file entry.')

    super(APFSContainerFileEntry, self).__init__(
        resolver_context, file_system, path_spec, is_root=is_root,
        is_virtual=is_virtual)
    self._name = None
    self._fsapfs_volume = fsapfs_volume

    if self._is_virtual:
      self.entry_type = definitions.FILE_ENTRY_TYPE_DIRECTORY
    else:
      self.entry_type = definitions.FILE_ENTRY_TYPE_FILE

  def _GetDirectory(self):
    """Retrieves a directory.

    Returns:
      APFSContainerDirectory: a directory.
    """
    if self.entry_type != definitions.FILE_ENTRY_TYPE_DIRECTORY:
      return None

    return apfs_container_directory.APFSContainerDirectory(
        self._file_system, self.path_spec)

  def _GetSubFileEntries(self):
    """Retrieves a sub file entries generator.

    Yields:
      APFSContainerFileEntry: a sub file entry.
    """
    if self._directory is None:
      self._directory = self._GetDirectory()

    if self._directory:
      for path_spec in self._directory.entries:
        yield APFSContainerFileEntry(
            self._resolver_context, self._file_system, path_spec)

  # TODO: expose date and time values.

  @property
  def name(self):
    """str: name of the file entry, which does not include the full path."""
    if self._name is None:
      location = getattr(self.path_spec, 'location', None)
      if location is not None:
        self._name = self._file_system.BasenamePath(location)
      else:
        volume_index = apfs_helper.APFSContainerPathSpecGetVolumeIndex(
            self.path_spec)
        if volume_index is not None:
          apfs_volume_index = volume_index + 1
          self._name = f'apfs{apfs_volume_index:d}'
        else:
          self._name = ''

    return self._name

  @property
  def size(self):
    """int: size of the file entry in bytes or None if not available."""
    if self._fsapfs_volume is None:
      return None

    # TODO: change libfsapfs so self._fsapfs_volume.size works
    return 0

  @property
  def sub_file_entries(self):
    """generator[APFSContainerFileEntry]: sub file entries."""
    return self._GetSubFileEntries()

  def GetAPFSVolume(self):
    """Retrieves an APFS volume.

    Returns:
      pyfsapfs.volume: an APFS volume or None if not available.
    """
    return self._fsapfs_volume

  def GetParentFileEntry(self):
    """Retrieves the parent file entry.

    Returns:
      APFSContainerFileEntry: parent file entry or None if not available.
    """
    volume_index = apfs_helper.APFSContainerPathSpecGetVolumeIndex(
        self.path_spec)
    if volume_index is None:
      return None

    return self._file_system.GetRootFileEntry()

  def IsLocked(self):
    """Determines if the file entry is locked.

    Returns:
      bool: True if the file entry is locked.
    """
    if self._fsapfs_volume is None:
      return False

    return self._fsapfs_volume.is_locked()

  def Unlock(self):
    """Unlocks the file entry.

    Returns:
      bool: True if the file entry was unlocked.
    """
    if not self._fsapfs_volume or not self._fsapfs_volume.is_locked():
      return True

    return apfs_helper.APFSUnlockVolume(
        self._fsapfs_volume, self.path_spec, resolver.Resolver.key_chain)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Tests for the File Allocation Table (FAT) file-like object."""

import unittest

from dfvfs.file_io import fat_file_io
from dfvfs.lib import definitions
from dfvfs.lib import errors
from dfvfs.path import factory as path_spec_factory
from dfvfs.resolver import context

from tests.file_io import test_lib


class FATFileTest(test_lib.FAT12ImageFileTestCase):
  """Tests the file-like object implementation using pyfsfat.file_entry."""

  _IDENTIFIER_ANOTHER_FILE = 0x62a0
  _IDENTIFIER_PASSWORDS_TXT = 0x1a80

  def setUp(self):
    """Sets up the needed objects used throughout the test."""
    super(FATFileTest, self).setUp()
    self._resolver_context = context.Context()
    test_path = self._GetTestFilePath(['fat12.raw'])
    self._SkipIfPathNotExists(test_path)

    test_os_path_spec = path_spec_factory.Factory.NewPathSpec(
        definitions.TYPE_INDICATOR_OS, location=test_path)
    self._raw_path_spec = path_spec_factory.Factory.NewPathSpec(
        definitions.TYPE_INDICATOR_RAW, parent=test_os_path_spec)

  def tearDown(self):
    """Cleans up the needed objects used throughout the test."""
    self._resolver_context.Empty()

  def testOpenCloseIdentifier(self):
    """Test the open and close functionality using an identifier."""
    path_spec = path_spec_factory.Factory.NewPathSpec(
        definitions.TYPE_INDICATOR_FAT,
        identifier=self._IDENTIFIER_PASSWORDS_TXT, parent=self._raw_path_spec)
    file_object = fat_file_io.FATFile(self._resolver_context, path_spec)

    self._TestOpenCloseIdentifier(file_object)

  def testOpenCloseLocation(self):
    """Test the open and close functionality using a location."""
    path_spec = path_spec_factory.Factory.NewPathSpec(
        definitions.TYPE_INDICATOR_FAT,
        identifier=self._IDENTIFIER_PASSWORDS_TXT, location='\\passwords.txt',
        parent=self._raw_path_spec)
    file_object = fat_file_io.FATFile(self._resolver_context, path_spec)

    self._TestOpenCloseLocation(file_object)

    # Try open with a path specification that has no parent.
    path_spec.parent = None
    file_object = fat_file_io.FATFile(self._resolver_context, path_spec)

    with self.assertRaises(errors.PathSpecError):
      self._TestOpenCloseLocation(file_object)

  def testSeek(self):
    """Test the seek functionality."""
    path_spec = path_spec_factory.Factory.NewPathSpec(
        definitions.TYPE_INDICATOR_FAT, location='\\a_directory\\another_file',
        identifier=self._IDENTIFIER_ANOTHER_FILE,
        parent=self._raw_path_spec)
    file_object = fat_file_io.FATFile(self._resolver_context, path_spec)

    self._TestSeek(file_object)

  def testRead(self):
    """Test the read functionality."""
    path_spec = path_spec_factory.Factory.NewPathSpec(
        definitions.TYPE_INDICATOR_FAT, location='\\passwords.txt',
        identifier=self._IDENTIFIER_PASSWORDS_TXT,
        parent=self._raw_path_spec)
    file_object = fat_file_io.FATFile(self._resolver_context, path_spec)

    self._TestRead(file_object)


if __name__ == '__main__':
  unittest.main()



# -*- coding: utf-8 -*-
"""The AES decrypter implementation."""

from cryptography.hazmat.primitives.ciphers import algorithms

from dfvfs.encryption import decrypter
from dfvfs.encryption import manager
from dfvfs.lib import definitions


class AESDecrypter(decrypter.CryptographyBlockCipherDecrypter):
  """AES decrypter using Cryptography."""

  ENCRYPTION_METHOD = definitions.ENCRYPTION_METHOD_AES

  _ENCRYPTION_MODES = frozenset([
      definitions.ENCRYPTION_MODE_CBC,
      definitions.ENCRYPTION_MODE_CFB,
      definitions.ENCRYPTION_MODE_ECB,
      definitions.ENCRYPTION_MODE_OFB])

  def __init__(
      self, cipher_mode=None, initialization_vector=None, key=None, **kwargs):
    """Initializes a decrypter.

    Args:
      cipher_mode (Optional[str]): cipher mode.
      initialization_vector (Optional[bytes]): initialization vector.
      key (Optional[bytes]): key.
      kwargs (dict): keyword arguments depending on the decrypter.

    Raises:
      ValueError: when key is not set or the cipher mode is not supported.
    """
    if not key:
      raise ValueError('Missing key.')

    if cipher_mode not in self._ENCRYPTION_MODES:
      raise ValueError(f'Unsupported cipher mode: {cipher_mode!s}')

    algorithm = algorithms.AES(key)

    super(AESDecrypter, self).__init__(
        algorithm=algorithm, cipher_mode=cipher_mode,
        initialization_vector=initialization_vector, **kwargs)


manager.EncryptionManager.RegisterDecrypter(AESDecrypter)



# -*- coding: utf-8 -*-
"""The APFS container directory implementation."""

from dfvfs.lib import apfs_helper
from dfvfs.path import apfs_container_path_spec
from dfvfs.vfs import directory


class APFSContainerDirectory(directory.Directory):
  """File system directory that uses pyfsapfs."""

  def _EntriesGenerator(self):
    """Retrieves directory entries.

    Since a directory can contain a vast number of entries using
    a generator is more memory efficient.

    Yields:
      APFSContainerPathSpec: a path specification.
    """
    # Only the virtual root file has directory entries.
    volume_index = apfs_helper.APFSContainerPathSpecGetVolumeIndex(
        self.path_spec)
    if volume_index is not None:
      return

    location = getattr(self.path_spec, 'location', None)
    if location is None or location != self._file_system.LOCATION_ROOT:
      return

    fsapfs_container = self._file_system.GetAPFSContainer()

    for volume_index in range(0, fsapfs_container.number_of_volumes):
      apfs_volume_index = volume_index + 1
      yield apfs_container_path_spec.APFSContainerPathSpec(
          location=f'/apfs{apfs_volume_index:d}', parent=self.path_spec.parent,
          volume_index=volume_index)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Tests for the VFS extent."""

import unittest

from dfvfs.vfs import extent

from tests import test_lib as shared_test_lib


class ExtentTest(shared_test_lib.BaseTestCase):
  """Tests the VFS extent."""

  def testInitialize(self):
    """Test the __init__ function."""
    test_extent = extent.Extent()
    self.assertIsNotNone(test_extent)


if __name__ == '__main__':
  unittest.main()



# -*- coding: utf-8 -*-
"""The EWF image path specification resolver helper implementation."""

from dfvfs.file_io import ewf_file_io
from dfvfs.lib import definitions
from dfvfs.resolver_helpers import manager
from dfvfs.resolver_helpers import resolver_helper


class EWFResolverHelper(resolver_helper.ResolverHelper):
  """EWF image resolver helper."""

  TYPE_INDICATOR = definitions.TYPE_INDICATOR_EWF

  def NewFileObject(self, resolver_context, path_spec):
    """Creates a new file input/output (IO) object.

    Args:
      resolver_context (Context): resolver context.
      path_spec (PathSpec): a path specification.

    Returns:
      FileIO: file input/output (IO) object.
    """
    return ewf_file_io.EWFFile(resolver_context, path_spec)


manager.ResolverHelperManager.RegisterHelper(EWFResolverHelper())



# -*- coding: utf-8 -*-
"""Shared test cases."""

from dfvfs.path import path_spec

from tests import test_lib as shared_test_lib


class TestPathSpec(path_spec.PathSpec):
  """Test path specification."""

  _IS_SYSTEM_LEVEL = True
  TYPE_INDICATOR = 'TEST'

  def __init__(self, **kwargs):
    """Initializes a test path specification."""
    super(TestPathSpec, self).__init__(parent=None, **kwargs)

  @property
  def comparable(self):
    """Comparable representation of the path specification."""
    return self._GetComparable()


class PathSpecTestCase(shared_test_lib.BaseTestCase):
  """The unit test case for path specification implementations."""

  def setUp(self):
    """Sets up the needed objects used throughout the test."""
    self._path_spec = TestPathSpec()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Tests for the file-like object implementation using pybde."""

import unittest

from dfvfs.file_io import tsk_file_io
from dfvfs.lib import definitions
from dfvfs.lib import errors
from dfvfs.path import factory as path_spec_factory
from dfvfs.resolver import context
from dfvfs.resolver import resolver

from tests.file_io import test_lib


class BDEFileWithKeyChainTest(test_lib.FAT12ImageFileTestCase):
  """Tests the BitLocker Drive Encryption (BDE) file-like object.

  The credentials are passed via the key chain.
  """

  _BDE_PASSWORD = 'bde-TEST'

  _IDENTIFIER_ANOTHER_FILE = 582
  _IDENTIFIER_PASSWORDS_TXT = 8

  def setUp(self):
    """Sets up the needed objects used throughout the test."""
    super(BDEFileWithKeyChainTest, self).setUp()
    self._resolver_context = context.Context()
    test_path = self._GetTestFilePath(['bdetogo.raw'])
    self._SkipIfPathNotExists(test_path)

    self._os_path_spec = path_spec_factory.Factory.NewPathSpec(
        definitions.TYPE_INDICATOR_OS, location=test_path)
    self._bde_path_spec = path_spec_factory.Factory.NewPathSpec(
        definitions.TYPE_INDICATOR_BDE, parent=self._os_path_spec)
    resolver.Resolver.key_chain.SetCredential(
        self._bde_path_spec, 'password', self._BDE_PASSWORD)

  def tearDown(self):
    """Cleans up the needed objects used throughout the test."""
    self._resolver_context.Empty()

  def testOpenCloseIdentifier(self):
    """Test the open and close functionality using an identifier."""
    path_spec = path_spec_factory.Factory.NewPathSpec(
        definitions.TYPE_INDICATOR_TSK, inode=self._IDENTIFIER_PASSWORDS_TXT,
        parent=self._bde_path_spec)
    file_object = tsk_file_io.TSKFile(self._resolver_context, path_spec)

    self._TestOpenCloseIdentifier(file_object)

  def testOpenCloseLocation(self):
    """Test the open and close functionality using a location."""
    path_spec = path_spec_factory.Factory.NewPathSpec(
        definitions.TYPE_INDICATOR_TSK, location='/passwords.txt',
        parent=self._bde_path_spec)
    file_object = tsk_file_io.TSKFile(self._resolver_context, path_spec)

    self._TestOpenCloseLocation(file_object)

    # Try open with a path specification that has no parent.
    path_spec.parent = None
    file_object = tsk_file_io.TSKFile(self._resolver_context, path_spec)

    with self.assertRaises(errors.PathSpecError):
      self._TestOpenCloseLocation(file_object)

  def testSeek(self):
    """Test the seek functionality."""
    path_spec = path_spec_factory.Factory.NewPathSpec(
        definitions.TYPE_INDICATOR_TSK, location='/a_directory/another_file',
        inode=self._IDENTIFIER_ANOTHER_FILE, parent=self._bde_path_spec)
    file_object = tsk_file_io.TSKFile(self._resolver_context, path_spec)

    self._TestSeek(file_object)

  def testRead(self):
    """Test the read functionality."""
    path_spec = path_spec_factory.Factory.NewPathSpec(
        definitions.TYPE_INDICATOR_TSK, location='/passwords.txt',
        inode=self._IDENTIFIER_PASSWORDS_TXT, parent=self._bde_path_spec)
    file_object = tsk_file_io.TSKFile(self._resolver_context, path_spec)

    self._TestRead(file_object)


class BDEFileWithPathSpecCredentialsTest(test_lib.FAT12ImageFileTestCase):
  """Tests the BitLocker Drive Encryption (BDE) file-like object.

  The credentials are passed via the path specification.
  """

  _BDE_PASSWORD = 'bde-TEST'

  _IDENTIFIER_ANOTHER_FILE = 582
  _IDENTIFIER_PASSWORDS_TXT = 8

  def setUp(self):
    """Sets up the needed objects used throughout the test."""
    super(BDEFileWithPathSpecCredentialsTest, self).setUp()
    self._resolver_context = context.Context()
    test_path = self._GetTestFilePath(['bdetogo.raw'])
    self._SkipIfPathNotExists(test_path)

    self._os_path_spec = path_spec_factory.Factory.NewPathSpec(
        definitions.TYPE_INDICATOR_OS, location=test_path)
    self._bde_path_spec = path_spec_factory.Factory.NewPathSpec(
        definitions.TYPE_INDICATOR_BDE, parent=self._os_path_spec,
        password=self._BDE_PASSWORD)

  def tearDown(self):
    """Cleans up the needed objects used throughout the test."""
    self._resolver_context.Empty()

  def testOpenCloseIdentifier(self):
    """Test the open and close functionality using an identifier."""
    path_spec = path_spec_factory.Factory.NewPathSpec(
        definitions.TYPE_INDICATOR_TSK, inode=self._IDENTIFIER_PASSWORDS_TXT,
        parent=self._bde_path_spec)
    file_object = tsk_file_io.TSKFile(self._resolver_context, path_spec)

    self._TestOpenCloseIdentifier(file_object)

  def testOpenCloseLocation(self):
    """Test the open and close functionality using a location."""
    path_spec = path_spec_factory.Factory.NewPathSpec(
        definitions.TYPE_INDICATOR_TSK, location='/passwords.txt',
        parent=self._bde_path_spec)
    file_object = tsk_file_io.TSKFile(self._resolver_context, path_spec)

    self._TestOpenCloseLocation(file_object)

    # Try open with a path specification that has no parent.
    path_spec.parent = None
    file_object = tsk_file_io.TSKFile(self._resolver_context, path_spec)

    with self.assertRaises(errors.PathSpecError):
      self._TestOpenCloseLocation(file_object)

  def testSeek(self):
    """Test the seek functionality."""
    path_spec = path_spec_factory.Factory.NewPathSpec(
        definitions.TYPE_INDICATOR_TSK, location='/a_directory/another_file',
        inode=self._IDENTIFIER_ANOTHER_FILE, parent=self._bde_path_spec)
    file_object = tsk_file_io.TSKFile(self._resolver_context, path_spec)

    self._TestSeek(file_object)

  def testRead(self):
    """Test the read functionality."""
    path_spec = path_spec_factory.Factory.NewPathSpec(
        definitions.TYPE_INDICATOR_TSK, location='/passwords.txt',
        inode=self._IDENTIFIER_PASSWORDS_TXT, parent=self._bde_path_spec)
    file_object = tsk_file_io.TSKFile(self._resolver_context, path_spec)

    self._TestRead(file_object)


if __name__ == '__main__':
  unittest.main()
